# pylint: disable=line-too-long

""" StructureToRust class for converting JSON Structure schema to Rust structs """

import json
import os
import re
from typing import Any, Dict, List, Set, Tuple, Union, Optional

from avrotize.common import pascal, snake, render_template

JsonNode = Dict[str, 'JsonNode'] | List['JsonNode'] | str | None

INDENT = '    '


class StructureToRust:
    """ Converts JSON Structure schema to Rust structs """

    def __init__(self, base_package: str = '', serde_annotation: bool = False) -> None:
        self.base_package = base_package.replace('.', '/').lower()
        self.serde_annotation = serde_annotation
        self.output_dir = os.getcwd()
        self.schema_doc: JsonNode = None
        self.generated_types_rust_package: Dict[str, str] = {}
        self.generated_structure_types: Dict[str, Dict[str, Union[str, Dict, List]]] = {}
        self.type_dict: Dict[str, Dict] = {}
        self.definitions: Dict[str, Any] = {}
        self.schema_registry: Dict[str, Dict] = {}

    reserved_words = [
        'as', 'break', 'const', 'continue', 'crate', 'else', 'enum', 'extern', 'false', 'fn', 'for', 'if', 'impl',
        'in', 'let', 'loop', 'match', 'mod', 'move', 'mut', 'pub', 'ref', 'return', 'self', 'Self', 'static',
        'struct', 'super', 'trait', 'true', 'type', 'unsafe', 'use', 'where', 'while', 'async', 'await', 'dyn',
    ]

    def safe_identifier(self, name: str) -> str:
        """Converts a name to a safe Rust identifier"""
        if name in StructureToRust.reserved_words:
            return f"{name}_"
        return name

    def escaped_identifier(self, name: str) -> str:
        """Converts a name to a safe Rust identifier with a leading r# prefix"""
        if name != "crate" and name in StructureToRust.reserved_words:
            return f"r#{name}"
        return name

    def safe_package(self, package: str) -> str:
        """Converts a package name to a safe Rust package name"""
        elements = package.split('::')
        return '::'.join([self.escaped_identifier(element) for element in elements])

    def get_qualified_name(self, namespace: str, name: str) -> str:
        """ Concatenates namespace and name with a dot separator """
        return f"{namespace}.{name}" if namespace != '' else name

    def sanitize_namespace(self, namespace: str) -> str:
        """Converts a namespace to a valid Rust module path by replacing dots with underscores"""
        return namespace.replace('.', '_')

    def concat_namespace(self, namespace: str, name: str) -> str:
        """ Concatenates namespace and name with a dot separator """
        if namespace and name:
            return f"{namespace}.{name}"
        elif namespace:
            return namespace
        else:
            return name

    def concat_package(self, package: str, name: str) -> str:
        """Concatenates package and name using a double colon separator"""
        return f"crate::{package.lower()}::{name.lower()}::{name}" if package else name

    def map_primitive_to_rust(self, structure_type: str, is_optional: bool = False) -> str:
        """ Maps JSON Structure primitive types to Rust types """
        optional_mapping = {
            'null': 'None',
            'boolean': 'Option<bool>',
            'string': 'Option<String>',
            'integer': 'Option<i32>',
            'number': 'Option<f64>',
            'int8': 'Option<i8>',
            'uint8': 'Option<u8>',
            'int16': 'Option<i16>',
            'uint16': 'Option<u16>',
            'int32': 'Option<i32>',
            'uint32': 'Option<u32>',
            'int64': 'Option<i64>',
            'uint64': 'Option<u64>',
            'int128': 'Option<i128>',
            'uint128': 'Option<u128>',
            'float8': 'Option<f32>',
            'float': 'Option<f32>',
            'double': 'Option<f64>',
            'binary32': 'Option<f32>',
            'binary64': 'Option<f64>',
            'decimal': 'Option<f64>',
            'binary': 'Option<Vec<u8>>',
            'date': 'Option<chrono::NaiveDate>',
            'time': 'Option<chrono::NaiveTime>',
            'datetime': 'Option<chrono::DateTime<chrono::Utc>>',
            'timestamp': 'Option<chrono::DateTime<chrono::Utc>>',
            'duration': 'Option<chrono::Duration>',
            'uuid': 'Option<uuid::Uuid>',
            'uri': 'Option<String>',
            'jsonpointer': 'Option<String>',
            'any': 'Option<serde_json::Value>',
        }
        required_mapping = {
            'null': 'None',
            'boolean': 'bool',
            'string': 'String',
            'integer': 'i32',
            'number': 'f64',
            'int8': 'i8',
            'uint8': 'u8',
            'int16': 'i16',
            'uint16': 'u16',
            'int32': 'i32',
            'uint32': 'u32',
            'int64': 'i64',
            'uint64': 'u64',
            'int128': 'i128',
            'uint128': 'u128',
            'float8': 'f32',
            'float': 'f32',
            'double': 'f64',
            'binary32': 'f32',
            'binary64': 'f64',
            'decimal': 'f64',
            'binary': 'Vec<u8>',
            'date': 'chrono::NaiveDate',
            'time': 'chrono::NaiveTime',
            'datetime': 'chrono::DateTime<chrono::Utc>',
            'timestamp': 'chrono::DateTime<chrono::Utc>',
            'duration': 'chrono::Duration',
            'uuid': 'uuid::Uuid',
            'uri': 'String',
            'jsonpointer': 'String',
            'any': 'serde_json::Value',
        }
        
        rust_fullname = structure_type
        if '.' in rust_fullname:
            type_name = pascal(structure_type.split('.')[-1])
            package_name = '::'.join(structure_type.split('.')[:-1]).lower()
            rust_fullname = self.safe_package(self.concat_package(package_name, type_name))
        
        if rust_fullname in self.generated_types_rust_package:
            return rust_fullname
        else:
            return required_mapping.get(structure_type, 'serde_json::Value') if not is_optional else optional_mapping.get(structure_type, 'Option<serde_json::Value>')

    def resolve_ref(self, ref: str, context_schema: Optional[Dict] = None) -> Optional[Dict]:
        """ Resolves a $ref to the actual schema definition """
        if not ref.startswith('#/'):
            if ref in self.schema_registry:
                return self.schema_registry[ref]
            return None

        path = ref[2:].split('/')
        schema = context_schema if context_schema else self.schema_doc
        for part in path:
            if not isinstance(schema, dict) or part not in schema:
                return None
            schema = schema[part]
        return schema

    def register_schema_ids(self, schema: Dict, base_uri: str = '') -> None:
        """ Recursively registers schemas with $id keywords """
        if not isinstance(schema, dict):
            return

        if '$id' in schema:
            schema_id = schema['$id']
            if base_uri and not schema_id.startswith(('http://', 'https://', 'urn:')):
                from urllib.parse import urljoin
                schema_id = urljoin(base_uri, schema_id)
            self.schema_registry[schema_id] = schema
            base_uri = schema_id

        if 'definitions' in schema:
            for def_name, def_schema in schema['definitions'].items():
                if isinstance(def_schema, dict):
                    self.register_schema_ids(def_schema, base_uri)

        if 'properties' in schema:
            for prop_name, prop_schema in schema['properties'].items():
                if isinstance(prop_schema, dict):
                    self.register_schema_ids(prop_schema, base_uri)

        for key in ['items', 'values', 'additionalProperties']:
            if key in schema and isinstance(schema[key], dict):
                self.register_schema_ids(schema[key], base_uri)

    def convert_structure_type_to_rust(self, class_name: str, field_name: str, structure_type: JsonNode, parent_namespace: str, nullable: bool = False) -> str:
        """ Converts JSON Structure type to Rust type """
        ns = self.sanitize_namespace(parent_namespace).replace('.', '::').lower()
        
        if isinstance(structure_type, str):
            return self.map_primitive_to_rust(structure_type, nullable)
        elif isinstance(structure_type, list):
            # Handle type unions
            non_null_types = [t for t in structure_type if t != 'null']
            has_null = 'null' in structure_type
            
            if len(non_null_types) == 1:
                inner_type = self.convert_structure_type_to_rust(class_name, field_name, non_null_types[0], parent_namespace, False)
                if has_null:
                    if inner_type.startswith('Option<'):
                        return inner_type
                    return f'Option<{inner_type}>'
                return inner_type
            else:
                # Multiple non-null types - generate a union enum
                return self.generate_union_enum(field_name, structure_type, parent_namespace)
        elif isinstance(structure_type, dict):
            # Handle $ref
            if '$ref' in structure_type:
                ref_schema = self.resolve_ref(structure_type['$ref'], self.schema_doc if isinstance(self.schema_doc, dict) else None)
                if ref_schema:
                    ref_path = structure_type['$ref'].split('/')
                    type_name = ref_path[-1]
                    ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_namespace
                    return self.generate_class_or_choice(ref_schema, ref_namespace, write_file=True, explicit_name=type_name)
                return 'serde_json::Value'

            # Handle enum keyword
            if 'enum' in structure_type:
                return self.generate_enum(structure_type, field_name, parent_namespace, write_file=True)

            # Handle type keyword
            if 'type' not in structure_type:
                return 'serde_json::Value'

            struct_type = structure_type['type']

            # Handle complex types
            if struct_type == 'object':
                return self.generate_class(structure_type, parent_namespace, write_file=True)
            elif struct_type == 'array':
                items_type = self.convert_structure_type_to_rust(
                    class_name, field_name+'Item', structure_type.get('items', {'type': 'any'}), parent_namespace)
                return f"Vec<{items_type}>"
            elif struct_type == 'set':
                items_type = self.convert_structure_type_to_rust(
                    class_name, field_name+'Item', structure_type.get('items', {'type': 'any'}), parent_namespace)
                return f"std::collections::HashSet<{items_type}>"
            elif struct_type == 'map':
                values_type = self.convert_structure_type_to_rust(
                    class_name, field_name+'Value', structure_type.get('values', {'type': 'any'}), parent_namespace)
                return f"std::collections::HashMap<String, {values_type}>"
            elif struct_type == 'choice':
                return self.generate_choice(structure_type, parent_namespace, write_file=True)
            elif struct_type == 'tuple':
                return self.generate_tuple(structure_type, parent_namespace, write_file=True)
            else:
                return self.convert_structure_type_to_rust(class_name, field_name, struct_type, parent_namespace, nullable)
        
        return 'serde_json::Value'

    def generate_class_or_choice(self, structure_schema: Dict, parent_namespace: str, write_file: bool = True, explicit_name: str = '') -> str:
        """ Generates a Class or Choice """
        struct_type = structure_schema.get('type', 'object')
        if struct_type == 'object':
            return self.generate_class(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        elif struct_type == 'choice':
            return self.generate_choice(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        elif struct_type == 'tuple':
            return self.generate_tuple(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        return 'serde_json::Value'

    def generate_class(self, structure_schema: Dict, parent_namespace: str, write_file: bool, explicit_name: str = '') -> str:
        """ Generates a Rust struct from JSON Structure object type """
        # Get name and namespace
        class_name = pascal(explicit_name if explicit_name else structure_schema.get('name', 'UnnamedClass'))
        schema_namespace = structure_schema.get('namespace', parent_namespace)
        namespace = self.sanitize_namespace(schema_namespace.lower())
        
        qualified_struct_name = self.safe_package(self.concat_package(namespace, class_name))
        if qualified_struct_name in self.generated_types_rust_package:
            return qualified_struct_name

        # Check if this is an abstract type
        is_abstract = structure_schema.get('abstract', False)

        # Handle inheritance ($extends)
        base_class = None
        if '$extends' in structure_schema:
            base_ref = structure_schema['$extends']
            if isinstance(self.schema_doc, dict):
                base_schema = self.resolve_ref(base_ref, self.schema_doc)
                if base_schema:
                    ref_path = base_ref.split('/')
                    base_name = ref_path[-1]
                    ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_namespace
                    base_class = self.generate_class(base_schema, ref_namespace, write_file=True, explicit_name=base_name)

        # Generate properties
        properties = structure_schema.get('properties', {})
        required_props = structure_schema.get('required', [])

        fields = []
        for prop_name, prop_schema in properties.items():
            # Skip const fields for now (they would need to be class constants)
            if 'const' in prop_schema:
                continue
            
            original_field_name = prop_name
            field_name = self.safe_identifier(snake(prop_name))
            
            # Determine if required
            is_required = prop_name in required_props if not isinstance(required_props, list) or \
                         len(required_props) == 0 or not isinstance(required_props[0], list) else \
                         any(prop_name in req_set for req_set in required_props)
            
            # Get property type
            prop_type = self.convert_structure_type_to_rust(class_name, field_name, prop_schema, schema_namespace, not is_required)
            
            # Add Option wrapper if not required and doesn't already have it
            if not is_required and not prop_type.startswith('Option<'):
                prop_type = f'Option<{prop_type}>'
            
            serde_rename = field_name != original_field_name
            
            fields.append({
                'original_name': original_field_name,
                'name': field_name,
                'type': prop_type,
                'serde_rename': serde_rename,
                'random_value': self.generate_random_value(prop_type)
            })

        # Get docstring
        doc = structure_schema.get('description', structure_schema.get('doc', class_name))

        # Prepare context for template
        context = {
            'serde_annotation': self.serde_annotation,
            'doc': doc,
            'struct_name': self.safe_identifier(class_name),
            'fields': fields,
            'is_abstract': is_abstract,
        }

        file_name = self.to_file_name(qualified_struct_name)
        target_file = os.path.join(self.output_dir, "src", file_name + ".rs")
        render_template('structuretorust/dataclass_struct.rs.jinja', target_file, **context)
        self.write_mod_rs(namespace)

        self.generated_types_rust_package[qualified_struct_name] = "struct"
        self.generated_structure_types[qualified_struct_name] = structure_schema

        return qualified_struct_name

    def generate_enum(self, structure_schema: Dict, field_name: str, parent_namespace: str, write_file: bool) -> str:
        """ Generates a Rust enum from JSON Structure enum keyword """
        enum_values = structure_schema.get('enum', [])
        if not enum_values:
            return 'serde_json::Value'

        # Determine enum name from field name
        enum_name = pascal(structure_schema.get('name', field_name + 'Enum'))
        schema_namespace = structure_schema.get('namespace', parent_namespace)
        namespace = self.sanitize_namespace(schema_namespace.lower())
        
        qualified_enum_name = self.safe_package(self.concat_package(namespace, enum_name))
        if qualified_enum_name in self.generated_types_rust_package:
            return qualified_enum_name

        # Convert enum values to valid Rust identifiers
        symbols = []
        for value in enum_values:
            if isinstance(value, str):
                # Convert to PascalCase and make it a valid Rust identifier
                symbol = pascal(value.replace('-', '_').replace(' ', '_'))
                symbols.append({'name': symbol, 'value': value})
            else:
                # For numeric values, use Value prefix
                symbols.append({'name': f"Value{value}", 'value': str(value)})

        doc = structure_schema.get('description', structure_schema.get('doc', enum_name))

        context = {
            'serde_annotation': self.serde_annotation,
            'enum_name': self.safe_identifier(enum_name),
            'symbols': symbols,
            'doc': doc,
        }

        file_name = self.to_file_name(qualified_enum_name)
        target_file = os.path.join(self.output_dir, "src", file_name + ".rs")
        render_template('structuretorust/dataclass_enum.rs.jinja', target_file, **context)
        self.write_mod_rs(namespace)

        self.generated_types_rust_package[qualified_enum_name] = "enum"
        self.generated_structure_types[qualified_enum_name] = structure_schema

        return qualified_enum_name

    def generate_choice(self, structure_schema: Dict, parent_namespace: str, write_file: bool, explicit_name: str = '') -> str:
        """ Generates a discriminated union (choice) type """
        choice_name = explicit_name if explicit_name else structure_schema.get('name', 'UnnamedChoice')
        schema_namespace = structure_schema.get('namespace', parent_namespace)
        namespace = self.sanitize_namespace(schema_namespace.lower())
        
        qualified_name = self.safe_package(self.concat_package(namespace, pascal(choice_name)))
        if qualified_name in self.generated_types_rust_package:
            return qualified_name

        choices = structure_schema.get('choices', {})
        
        # Generate types for each choice
        choice_types = []
        for choice_key, choice_schema in choices.items():
            if isinstance(choice_schema, dict):
                if '$ref' in choice_schema:
                    # Resolve reference and generate the type
                    ref_schema = self.resolve_ref(choice_schema['$ref'], self.schema_doc if isinstance(self.schema_doc, dict) else None)
                    if ref_schema:
                        ref_path = choice_schema['$ref'].split('/')
                        ref_name = ref_path[-1]
                        ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_namespace
                        type_ref = self.generate_class(ref_schema, ref_namespace, write_file=True, explicit_name=ref_name)
                        type_name = type_ref.split('::')[-1]
                        choice_types.append({
                            'variant_name': pascal(choice_key),
                            'type': type_name,
                            'tag': choice_key
                        })
                elif 'type' in choice_schema:
                    # Generate inline type
                    rust_type = self.convert_structure_type_to_rust(choice_name, choice_key, choice_schema, schema_namespace)
                    choice_types.append({
                        'variant_name': pascal(choice_key),
                        'type': rust_type,
                        'tag': choice_key
                    })

        doc = structure_schema.get('description', structure_schema.get('doc', choice_name))

        context = {
            'serde_annotation': self.serde_annotation,
            'union_enum_name': self.safe_identifier(pascal(choice_name)),
            'variants': choice_types,
            'doc': doc,
        }

        file_name = self.to_file_name(qualified_name)
        target_file = os.path.join(self.output_dir, "src", file_name + ".rs")
        render_template('structuretorust/dataclass_union.rs.jinja', target_file, **context)
        self.write_mod_rs(namespace)

        self.generated_types_rust_package[qualified_name] = "choice"
        self.generated_structure_types[qualified_name] = structure_schema

        return qualified_name

    def generate_tuple(self, structure_schema: Dict, parent_namespace: str, write_file: bool, explicit_name: str = '') -> str:
        """ Generates a Rust tuple type from JSON Structure tuple """
        # For tuples, we generate a struct with numbered fields
        tuple_name = explicit_name if explicit_name else structure_schema.get('name', 'UnnamedTuple')
        return self.generate_class(structure_schema, parent_namespace, write_file, explicit_name=tuple_name)

    def generate_union_enum(self, field_name: str, structure_type: List, namespace: str) -> str:
        """Generates a union enum for Rust"""
        ns = self.sanitize_namespace(namespace.replace('.', '::').lower())
        union_enum_name = pascal(field_name) + 'Union'
        
        non_null_types = [t for t in structure_type if t != 'null']
        has_null = 'null' in structure_type
        
        union_types = []
        for i, t in enumerate(non_null_types):
            type_name = self.convert_structure_type_to_rust(field_name + "Option" + str(i), field_name + "Option" + str(i), t, namespace)
            variant_name = pascal(type_name.rsplit('::',1)[-1].replace('<', '').replace('>', '').replace(',', ''))
            union_types.append({
                'variant_name': variant_name,
                'type': type_name,
            })
        
        qualified_union_enum_name = self.safe_package(self.concat_package(ns, union_enum_name))
        
        context = {
            'serde_annotation': self.serde_annotation,
            'union_enum_name': union_enum_name,
            'variants': union_types,
            'doc': f'Union type for {field_name}',
        }

        file_name = self.to_file_name(qualified_union_enum_name)
        target_file = os.path.join(self.output_dir, "src", file_name + ".rs").lower()
        render_template('structuretorust/dataclass_union.rs.jinja', target_file, **context)
        self.generated_types_rust_package[qualified_union_enum_name] = "union"
        self.write_mod_rs(namespace)

        return qualified_union_enum_name

    def to_file_name(self, qualified_name):
        """Converts a qualified union enum name to a file name"""
        if qualified_name.startswith('crate::'):
            qualified_name = qualified_name[(len('crate::')):]
        qualified_name = qualified_name.replace('r#', '')
        return qualified_name.rsplit('::',1)[0].replace('::', os.sep).lower()

    def generate_random_value(self, rust_type: str) -> str:
        """Generates a random value for a given Rust type"""
        if rust_type == 'String' or rust_type == 'Option<String>':
            return 'format!("random_string_{}", rand::Rng::gen::<u32>(&mut rng))'
        elif rust_type == 'bool' or rust_type == 'Option<bool>':
            return 'rand::Rng::gen::<bool>(&mut rng)'
        elif rust_type in ['i8', 'Option<i8>']:
            return 'rand::Rng::gen_range(&mut rng, 1..101) as i8'
        elif rust_type in ['u8', 'Option<u8>']:
            return 'rand::Rng::gen_range(&mut rng, 1..101) as u8'
        elif rust_type in ['i16', 'Option<i16>']:
            return 'rand::Rng::gen_range(&mut rng, 1..101) as i16'
        elif rust_type in ['u16', 'Option<u16>']:
            return 'rand::Rng::gen_range(&mut rng, 1..101) as u16'
        elif rust_type in ['i32', 'Option<i32>']:
            return 'rand::Rng::gen_range(&mut rng, 1..101)'
        elif rust_type in ['u32', 'Option<u32>']:
            return 'rand::Rng::gen_range(&mut rng, 1..101) as u32'
        elif rust_type in ['i64', 'Option<i64>']:
            return 'rand::Rng::gen_range(&mut rng, 1..101) as i64'
        elif rust_type in ['u64', 'Option<u64>']:
            return 'rand::Rng::gen_range(&mut rng, 1..101) as u64'
        elif rust_type in ['i128', 'Option<i128>']:
            return 'rand::Rng::gen_range(&mut rng, 1..101) as i128'
        elif rust_type in ['u128', 'Option<u128>']:
            return 'rand::Rng::gen_range(&mut rng, 1..101) as u128'
        elif rust_type in ['f32', 'Option<f32>']:
            return '(rand::Rng::gen::<f32>(&mut rng)*1000.0).round()/1000.0'
        elif rust_type in ['f64', 'Option<f64>']:
            return '(rand::Rng::gen::<f64>(&mut rng)*1000.0).round()/1000.0'
        elif rust_type == 'Vec<u8>' or rust_type == 'Option<Vec<u8>>':
            return 'vec![rand::Rng::gen::<u8>(&mut rng); 10]'
        elif rust_type == 'chrono::NaiveDate' or rust_type == 'Option<chrono::NaiveDate>':
            return 'chrono::NaiveDate::from_ymd_opt(rand::Rng::gen_range(&mut rng, 2000..2023), rand::Rng::gen_range(&mut rng, 1..13), rand::Rng::gen_range(&mut rng, 1..29)).unwrap()'
        elif rust_type == 'chrono::NaiveTime' or rust_type == 'Option<chrono::NaiveTime>':
            return 'chrono::NaiveTime::from_hms_opt(rand::Rng::gen_range(&mut rng, 0..24),rand::Rng::gen_range(&mut rng, 0..60), rand::Rng::gen_range(&mut rng, 0..60)).unwrap()'
        elif rust_type == 'chrono::DateTime<chrono::Utc>' or rust_type == 'Option<chrono::DateTime<chrono::Utc>>':
            return 'chrono::Utc::now()'
        elif rust_type == 'chrono::Duration' or rust_type == 'Option<chrono::Duration>':
            return 'chrono::Duration::seconds(rand::Rng::gen_range(&mut rng, 0..86400))'
        elif rust_type == 'uuid::Uuid' or rust_type == 'Option<uuid::Uuid>':
            return 'uuid::Uuid::new_v4()'
        elif rust_type.startswith('std::collections::HashMap<String, '):
            inner_type = rust_type.split(', ')[1][:-1]
            return f'(0..3).map(|_| (format!("key_{{}}", rand::Rng::gen::<u32>(&mut rng)), {self.generate_random_value(inner_type)})).collect()'
        elif rust_type.startswith('std::collections::HashSet<'):
            inner_type = rust_type[27:-1]
            return f'(0..3).map(|_| {self.generate_random_value(inner_type)}).collect()'
        elif rust_type.startswith('Vec<'):
            inner_type = rust_type[4:-1]
            return f'(0..3).map(|_| {self.generate_random_value(inner_type)}).collect()'
        elif rust_type in self.generated_types_rust_package:
            return f'{rust_type}::generate_random_instance()'
        else:
            return 'Default::default()'

    def write_mod_rs(self, namespace: str):
        """Writes the mod.rs file for a Rust module"""
        # Sanitize namespace to replace dots with underscores
        sanitized_namespace = self.sanitize_namespace(namespace)
        directories = sanitized_namespace.split('.')
        for i in range(len(directories)):
            sub_package = '::'.join(directories[:i + 1])
            directory_path = os.path.join(
                self.output_dir, "src", sub_package.replace('.', os.sep).replace('::', os.sep))
            if not os.path.exists(directory_path):
                os.makedirs(directory_path, exist_ok=True)
            mod_rs_path = os.path.join(directory_path, "mod.rs")
            
            types = [file.replace('.rs', '') for file in os.listdir(directory_path) if file.endswith('.rs') and file != "mod.rs"]
            mod_statements = '\n'.join(f'pub mod {self.escaped_identifier(typ.lower())};' for typ in types)
            mods = [dir for dir in os.listdir(directory_path) if os.path.isdir(os.path.join(directory_path, dir))]
            mod_statements += '\n' + '\n'.join(f'pub mod {self.escaped_identifier(mod.lower())};' for mod in mods)

            with open(mod_rs_path, 'w', encoding='utf-8') as file:
                file.write(mod_statements)

    def write_cargo_toml(self):
        """Writes the Cargo.toml file for the Rust project"""
        dependencies = []
        if self.serde_annotation:
            dependencies.append('serde = { version = "1.0", features = ["derive"] }')
            dependencies.append('serde_json = "1.0"')
        dependencies.append('chrono = { version = "0.4", features = ["serde"] }')
        dependencies.append('uuid = { version = "1.11", features = ["serde", "v4"] }')
        dependencies.append('flate2 = "1.0"')
        dependencies.append('rand = "0.8"')

        cargo_toml_content =  f"[package]\n"
        cargo_toml_content += f"name = \"{self.base_package.replace('/', '_')}\"\n"
        cargo_toml_content += f"version = \"0.1.0\"\n"
        cargo_toml_content += f"edition = \"2021\"\n\n"
        cargo_toml_content += f"[dependencies]\n"
        cargo_toml_content += "\n".join(f"{dependency}" for dependency in dependencies)
        cargo_toml_path = os.path.join(self.output_dir, "Cargo.toml")
        with open(cargo_toml_path, 'w', encoding='utf-8') as file:
            file.write(cargo_toml_content)

    def write_lib_rs(self):
        """Writes the lib.rs file for the Rust project"""
        modules = {name[(len('crate::')):].split('::')[0].replace('.', '_') for name in self.generated_types_rust_package if name.startswith('crate::')}
        mod_statements = '\n'.join(f'pub mod {self.escaped_identifier(module)};' for module in sorted(modules))
        
        lib_rs_content = f"""
// This is the library entry point

{mod_statements}
"""
        lib_rs_path = os.path.join(self.output_dir, "src", "lib.rs")
        if not os.path.exists(os.path.dirname(lib_rs_path)):
            os.makedirs(os.path.dirname(lib_rs_path), exist_ok=True)
        with open(lib_rs_path, 'w', encoding='utf-8') as file:
            file.write(lib_rs_content)

    def process_definitions(self, definitions: Dict, namespace_path: str) -> None:
        """ Processes the definitions section recursively """
        for name, definition in definitions.items():
            if isinstance(definition, dict):
                if 'type' in definition:
                    # This is a type definition
                    current_namespace = self.concat_namespace(namespace_path, '')
                    check_namespace = self.sanitize_namespace(current_namespace.lower())
                    check_name = pascal(name)
                    check_ref = self.safe_package(self.concat_package(check_namespace, check_name))
                    if check_ref not in self.generated_types_rust_package:
                        self.generate_class_or_choice(definition, current_namespace, write_file=True, explicit_name=name)
                else:
                    # This is a namespace
                    new_namespace = self.concat_namespace(namespace_path, name)
                    self.process_definitions(definition, new_namespace)

    def convert_schema(self, schema: JsonNode, output_dir: str):
        """Converts JSON Structure schema to Rust"""
        if not isinstance(schema, list):
            schema = [schema]
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)
        self.output_dir = output_dir

        # Register all schema IDs first
        for structure_schema in (s for s in schema if isinstance(s, dict)):
            self.register_schema_ids(structure_schema)

        # Process each schema
        for structure_schema in (s for s in schema if isinstance(s, dict)):
            self.schema_doc = structure_schema
            
            # Store definitions for later use
            if 'definitions' in structure_schema:
                self.definitions = structure_schema['definitions']

            # Process root type FIRST
            if 'type' in structure_schema:
                self.generate_class_or_choice(structure_schema, '', write_file=True)
            elif '$root' in structure_schema:
                root_ref = structure_schema['$root']
                root_schema = self.resolve_ref(root_ref, structure_schema)
                if root_schema:
                    ref_path = root_ref.split('/')
                    type_name = ref_path[-1]
                    ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else ''
                    self.generate_class_or_choice(root_schema, ref_namespace, write_file=True, explicit_name=type_name)

            # Process definitions
            if 'definitions' in structure_schema:
                self.process_definitions(self.definitions, '')

        self.write_cargo_toml()
        self.write_lib_rs()

    def convert(self, structure_schema_path: str, output_dir: str):
        """Converts JSON Structure schema to Rust"""
        with open(structure_schema_path, 'r', encoding='utf-8') as file:
            schema = json.load(file)
        self.convert_schema(schema, output_dir)


def convert_structure_to_rust(structure_schema_path: str, rust_file_path: str, package_name: str = '', serde_annotation: bool = False):
    """Converts JSON Structure schema to Rust structs

    Args:
        structure_schema_path (str): JSON Structure input schema path
        rust_file_path (str): Output Rust file path
        package_name (str): Base package name
        serde_annotation (bool): Include Serde annotations
    """
    if not package_name:
        package_name = os.path.splitext(os.path.basename(structure_schema_path))[0].lower().replace('-', '_')

    structtorust = StructureToRust()
    structtorust.base_package = package_name
    structtorust.serde_annotation = serde_annotation
    structtorust.convert(structure_schema_path, rust_file_path)


def convert_structure_schema_to_rust(structure_schema: JsonNode, output_dir: str, package_name: str = '', serde_annotation: bool = False):
    """Converts JSON Structure schema to Rust structs

    Args:
        structure_schema (JsonNode): JSON Structure schema as a dictionary or list of dictionaries
        output_dir (str): Output directory path
        package_name (str): Base package name
        serde_annotation (bool): Include Serde annotations
    """
    structtorust = StructureToRust()
    structtorust.base_package = package_name
    structtorust.serde_annotation = serde_annotation
    structtorust.convert_schema(structure_schema, output_dir)



# pylint: disable=line-too-long

""" StructureToTypeScript class for converting JSON Structure schema to TypeScript classes """

import json
import os
import random
import re
from typing import Any, Dict, List, Set, Tuple, Union, Optional

from avrotize.common import pascal, process_template

JsonNode = Dict[str, 'JsonNode'] | List['JsonNode'] | str | None

INDENT = '    '


def is_typescript_reserved_word(word: str) -> bool:
    """Checks if a word is a TypeScript reserved word"""
    reserved_words = [
        'break', 'case', 'catch', 'class', 'const', 'continue', 'debugger',
        'default', 'delete', 'do', 'else', 'export', 'extends', 'finally',
        'for', 'function', 'if', 'import', 'in', 'instanceof', 'new', 'return',
        'super', 'switch', 'this', 'throw', 'try', 'typeof', 'var', 'void',
        'while', 'with', 'yield', 'enum', 'string', 'number', 'boolean', 'symbol',
        'type', 'namespace', 'module', 'declare', 'abstract', 'readonly',
    ]
    return word in reserved_words


class StructureToTypeScript:
    """ Converts JSON Structure schema to TypeScript classes """

    def __init__(self, base_package: str = '', typedjson_annotation=False, avro_annotation=False) -> None:
        self.base_package = base_package
        self.typedjson_annotation = typedjson_annotation
        self.avro_annotation = avro_annotation
        self.output_dir = os.getcwd()
        self.schema_doc: JsonNode = None
        self.generated_types: Dict[str, str] = {}
        self.generated_structure_types: Dict[str, Dict[str, Union[str, Dict, List]]] = {}
        self.type_dict: Dict[str, Dict] = {}
        self.definitions: Dict[str, Any] = {}
        self.schema_registry: Dict[str, Dict] = {}

    def get_qualified_name(self, namespace: str, name: str) -> str:
        """ Concatenates namespace and name with a dot separator """
        return f"{namespace}.{name}" if namespace != '' else name

    def concat_namespace(self, namespace: str, name: str) -> str:
        """ Concatenates namespace and name with a dot separator """
        if namespace and name:
            return f"{namespace}.{name}"
        elif namespace:
            return namespace
        else:
            return name

    def map_primitive_to_typescript(self, structure_type: str) -> str:
        """ Maps JSON Structure primitive types to TypeScript types """
        mapping = {
            'null': 'null',
            'boolean': 'boolean',
            'string': 'string',
            'integer': 'number',
            'number': 'number',
            'int8': 'number',
            'uint8': 'number',
            'int16': 'number',
            'uint16': 'number',
            'int32': 'number',
            'uint32': 'number',
            'int64': 'number',
            'uint64': 'number',
            'int128': 'bigint',
            'uint128': 'bigint',
            'float8': 'number',
            'float': 'number',
            'double': 'number',
            'binary32': 'number',
            'binary64': 'number',
            'decimal': 'string',
            'binary': 'string',
            'bytes': 'string',
            'date': 'Date',
            'time': 'Date',
            'datetime': 'Date',
            'timestamp': 'Date',
            'duration': 'string',
            'uuid': 'string',
            'uri': 'string',
            'jsonpointer': 'string',
            'any': 'any'
        }
        qualified_class_name = self.get_qualified_name(
            self.base_package.lower(), structure_type.lower())
        if qualified_class_name in self.generated_types:
            result = qualified_class_name
        else:
            result = mapping.get(structure_type, 'any')
        return result

    def is_typescript_primitive(self, type_name: str) -> bool:
        """ Checks if a type is a TypeScript primitive type """
        return type_name in ['null', 'boolean', 'number', 'bigint', 'string', 'Date', 'any']

    def safe_name(self, name: str) -> str:
        """Converts a name to a safe TypeScript name"""
        if is_typescript_reserved_word(name):
            return name + "_"
        return name

    def pascal_type_name(self, ref: str) -> str:
        """Converts a reference to a type name"""
        return '_'.join([pascal(part) for part in ref.split('.')[-1].split('_')])

    def typescript_package_from_structure_type(self, namespace: str, type_name: str) -> str:
        """Gets the TypeScript package from a type name"""
        if '.' in type_name:
            # Type name contains dots, use it as package path
            type_name_package = '.'.join([part.lower() for part in type_name.split('.')])
            package = type_name_package
        else:
            # Use namespace as package, don't add type name to package
            namespace_package = '.'.join([part.lower() for part in namespace.split('.')]) if namespace else ''
            package = namespace_package
        if self.base_package:
            package = self.base_package + ('.' if package else '') + package
        return package

    def typescript_type_from_structure_type(self, type_name: str) -> str:
        """Gets the TypeScript class from a type name"""
        return self.pascal_type_name(type_name)

    def typescript_fully_qualified_name_from_structure_type(self, namespace: str, type_name: str) -> str:
        """Gets the fully qualified TypeScript class name from a Structure type."""
        package = self.typescript_package_from_structure_type(namespace, type_name)
        return package + ('.' if package else '') + self.typescript_type_from_structure_type(type_name)

    def strip_package_from_fully_qualified_name(self, fully_qualified_name: str) -> str:
        """Strips the package from a fully qualified name"""
        return fully_qualified_name.split('.')[-1]

    def strip_nullable(self, ts_type: str) -> str:
        """Strip nullable marker from TypeScript type"""
        if ts_type.endswith(' | null') or ts_type.endswith('| null'):
            return ts_type.replace(' | null', '').replace('| null', '')
        if ts_type.endswith('| undefined') or ts_type.endswith(' | undefined'):
            return ts_type.replace(' | undefined', '').replace('| undefined', '')
        return ts_type

    def resolve_ref(self, ref: str, context_schema: Optional[Dict] = None) -> Optional[Dict]:
        """ Resolves a $ref to the actual schema definition """
        if not ref.startswith('#/'):
            if ref in self.schema_registry:
                return self.schema_registry[ref]
            return None

        path = ref[2:].split('/')
        schema = context_schema if context_schema else self.schema_doc
        for part in path:
            if not isinstance(schema, dict) or part not in schema:
                return None
            schema = schema[part]
        return schema

    def register_schema_ids(self, schema: Dict, base_uri: str = '') -> None:
        """ Recursively registers schemas with $id keywords """
        if not isinstance(schema, dict):
            return

        if '$id' in schema:
            schema_id = schema['$id']
            if base_uri and not schema_id.startswith(('http://', 'https://', 'urn:')):
                from urllib.parse import urljoin
                schema_id = urljoin(base_uri, schema_id)
            self.schema_registry[schema_id] = schema
            base_uri = schema_id

        if 'definitions' in schema:
            for def_name, def_schema in schema['definitions'].items():
                if isinstance(def_schema, dict):
                    self.register_schema_ids(def_schema, base_uri)

        if 'properties' in schema:
            for prop_name, prop_schema in schema['properties'].items():
                if isinstance(prop_schema, dict):
                    self.register_schema_ids(prop_schema, base_uri)

        for key in ['items', 'values', 'additionalProperties']:
            if key in schema and isinstance(schema[key], dict):
                self.register_schema_ids(schema[key], base_uri)

    def convert_structure_type_to_typescript(self, class_name: str, field_name: str, 
                                            structure_type: JsonNode, parent_namespace: str, 
                                            import_types: Set[str]) -> str:
        """ Converts JSON Structure type to TypeScript type """
        if isinstance(structure_type, str):
            ts_type = self.map_primitive_to_typescript(structure_type)
            return ts_type
        elif isinstance(structure_type, list):
            # Handle type unions
            non_null_types = [t for t in structure_type if t != 'null']
            if len(non_null_types) == 1:
                inner_type = self.convert_structure_type_to_typescript(
                    class_name, field_name, non_null_types[0], parent_namespace, import_types)
                if 'null' in structure_type:
                    return f'{inner_type} | null'
                return inner_type
            else:
                union_types = [self.convert_structure_type_to_typescript(
                    class_name, field_name, t, parent_namespace, import_types) for t in non_null_types]
                result = ' | '.join(union_types)
                if 'null' in structure_type:
                    result += ' | null'
                return result
        elif isinstance(structure_type, dict):
            # Handle $ref
            if '$ref' in structure_type:
                ref_schema = self.resolve_ref(structure_type['$ref'], self.schema_doc)
                if ref_schema:
                    ref_path = structure_type['$ref'].split('/')
                    type_name = ref_path[-1]
                    ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_namespace
                    ref = self.generate_class_or_choice(ref_schema, ref_namespace, write_file=True, explicit_name=type_name)
                    import_types.add(ref)
                    return self.strip_package_from_fully_qualified_name(ref)
                return 'any'

            # Handle enum keyword
            if 'enum' in structure_type:
                enum_ref = self.generate_enum(structure_type, field_name, parent_namespace, write_file=True)
                import_types.add(enum_ref)
                return self.strip_package_from_fully_qualified_name(enum_ref)

            # Handle type keyword
            if 'type' not in structure_type:
                return 'any'

            struct_type = structure_type['type']

            # Handle complex types
            if struct_type == 'object':
                class_ref = self.generate_class(structure_type, parent_namespace, write_file=True)
                import_types.add(class_ref)
                return self.strip_package_from_fully_qualified_name(class_ref)
            elif struct_type == 'array':
                items_type = self.convert_structure_type_to_typescript(
                    class_name, field_name+'Array', structure_type.get('items', {'type': 'any'}), 
                    parent_namespace, import_types)
                return f"{items_type}[]"
            elif struct_type == 'set':
                items_type = self.convert_structure_type_to_typescript(
                    class_name, field_name+'Set', structure_type.get('items', {'type': 'any'}), 
                    parent_namespace, import_types)
                return f"Set<{items_type}>"
            elif struct_type == 'map':
                values_type = self.convert_structure_type_to_typescript(
                    class_name, field_name+'Map', structure_type.get('values', {'type': 'any'}), 
                    parent_namespace, import_types)
                return f"{{ [key: string]: {values_type} }}"
            elif struct_type == 'choice':
                # Generate choice returns a Union type and populates import_types with the choice types
                return self.generate_choice(structure_type, parent_namespace, write_file=True, import_types=import_types)
            elif struct_type == 'tuple':
                tuple_ref = self.generate_tuple(structure_type, parent_namespace, write_file=True)
                import_types.add(tuple_ref)
                return self.strip_package_from_fully_qualified_name(tuple_ref)
            else:
                return self.convert_structure_type_to_typescript(class_name, field_name, struct_type, parent_namespace, import_types)
        return 'any'

    def generate_class_or_choice(self, structure_schema: Dict, parent_namespace: str, 
                                 write_file: bool = True, explicit_name: str = '') -> str:
        """ Generates a Class or Choice """
        struct_type = structure_schema.get('type', 'object')
        if struct_type == 'object':
            return self.generate_class(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        elif struct_type == 'choice':
            return self.generate_choice(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        elif struct_type == 'tuple':
            return self.generate_tuple(structure_schema, parent_namespace, write_file, explicit_name=explicit_name)
        return 'any'

    def generate_class(self, structure_schema: Dict, parent_namespace: str, 
                      write_file: bool, explicit_name: str = '') -> str:
        """ Generates a TypeScript class/interface from JSON Structure object type """
        import_types: Set[str] = set()

        # Get name and namespace
        class_name = pascal(explicit_name if explicit_name else structure_schema.get('name', 'UnnamedClass'))
        schema_namespace = structure_schema.get('namespace', parent_namespace)
        namespace = self.concat_namespace(self.base_package, schema_namespace).lower()
        typescript_qualified_name = self.typescript_fully_qualified_name_from_structure_type(schema_namespace, class_name)
        
        if typescript_qualified_name in self.generated_types:
            return typescript_qualified_name

        # Check if this is an abstract type
        is_abstract = structure_schema.get('abstract', False)

        # Handle inheritance ($extends)
        base_class = None
        if '$extends' in structure_schema:
            base_ref = structure_schema['$extends']
            if isinstance(self.schema_doc, dict):
                base_schema = self.resolve_ref(base_ref, self.schema_doc)
                if base_schema:
                    base_class_ref = self.generate_class(base_schema, parent_namespace, write_file=True)
                    base_class = self.strip_package_from_fully_qualified_name(base_class_ref)
                    import_types.add(base_class_ref)

        # Collect properties
        properties = structure_schema.get('properties', {})
        required_props = set(structure_schema.get('required', []))

        # Handle add-ins ($uses)
        if '$uses' in structure_schema and isinstance(structure_schema['$uses'], list):
            for addin_ref in structure_schema['$uses']:
                if isinstance(addin_ref, str):
                    # Resolve the add-in reference
                    addin_schema = self.resolve_ref(addin_ref, self.schema_doc)
                    if addin_schema and 'properties' in addin_schema:
                        properties.update(addin_schema['properties'])
                        if 'required' in addin_schema:
                            required_props.update(addin_schema['required'])

        # Generate fields
        fields = []
        for prop_name, prop_schema in properties.items():
            field_type = self.convert_structure_type_to_typescript(
                class_name, prop_name, prop_schema, namespace, import_types)
            is_required = prop_name in required_props
            is_optional = not is_required
            
            fields.append({
                'name': self.safe_name(prop_name),
                'original_name': prop_name,
                'type': field_type,
                'type_no_null': self.strip_nullable(field_type),
                'is_required': is_required,
                'is_optional': is_optional,
                'is_primitive': self.is_typescript_primitive(self.strip_nullable(field_type).replace('[]', '')),
                'docstring': prop_schema.get('description', '') if isinstance(prop_schema, dict) else ''
            })

        # Build imports
        imports_with_paths: Dict[str, str] = {}
        for import_type in import_types:
            if import_type == typescript_qualified_name:
                continue
            import_is_enum = import_type in self.generated_types and self.generated_types[import_type] == 'enum'
            import_type_parts = import_type.split('.')
            import_type_name = pascal(import_type_parts[-1])
            import_path = '/'.join(import_type_parts)
            current_path = '/'.join(namespace.split('.'))
            relative_import_path = os.path.relpath(import_path, current_path).replace(os.sep, '/')
            if not relative_import_path.startswith('.'):
                relative_import_path = f'./{relative_import_path}'
            imports_with_paths[import_type_name] = relative_import_path + '.js'

        # Generate class definition using template
        class_definition = process_template(
            "structuretots/class_core.ts.jinja",
            namespace=namespace,
            class_name=class_name,
            base_class=base_class,
            is_abstract=is_abstract,
            docstring=structure_schema.get('description', '').strip() if 'description' in structure_schema else f'A {class_name} class.',
            fields=fields,
            imports=imports_with_paths,
            typedjson_annotation=self.typedjson_annotation,
        )

        if write_file:
            self.write_to_file(namespace, class_name, class_definition)
            # Generate test class
            if not is_abstract:  # Don't generate tests for abstract classes
                self.generate_test_class(namespace, class_name, fields)
        self.generated_types[typescript_qualified_name] = 'class'
        return typescript_qualified_name

    def generate_enum(self, structure_schema: Dict, field_name: str, parent_namespace: str, 
                     write_file: bool = True) -> str:
        """ Generates a TypeScript enum from JSON Structure enum """
        enum_name = pascal(structure_schema.get('name', field_name + 'Enum'))
        namespace = self.concat_namespace(self.base_package, structure_schema.get('namespace', parent_namespace)).lower()
        typescript_qualified_name = self.typescript_fully_qualified_name_from_structure_type(parent_namespace, enum_name)
        
        if typescript_qualified_name in self.generated_types:
            return typescript_qualified_name

        symbols = structure_schema.get('enum', [])
        
        enum_definition = process_template(
            "structuretots/enum_core.ts.jinja",
            namespace=namespace,
            enum_name=enum_name,
            docstring=structure_schema.get('description', '').strip() if 'description' in structure_schema else f'A {enum_name} enum.',
            symbols=symbols,
        )

        if write_file:
            self.write_to_file(namespace, enum_name, enum_definition)
        self.generated_types[typescript_qualified_name] = 'enum'
        return typescript_qualified_name

    def generate_choice(self, structure_schema: Dict, parent_namespace: str, 
                       write_file: bool = True, explicit_name: str = '', 
                       import_types: Optional[Set[str]] = None) -> str:
        """ Generates a TypeScript union type from JSON Structure choice type """
        if import_types is None:
            import_types = set()
        
        choice_name = pascal(explicit_name if explicit_name else structure_schema.get('name', 'Choice'))
        schema_namespace = structure_schema.get('namespace', parent_namespace)
        
        # If the choice extends a base class, generate the base class first
        if '$extends' in structure_schema:
            base_ref = structure_schema['$extends']
            if isinstance(self.schema_doc, dict):
                base_schema = self.resolve_ref(base_ref, self.schema_doc)
                if base_schema:
                    # Generate the base class
                    ref_path = base_ref.split('/')
                    base_name = ref_path[-1]
                    ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_namespace
                    self.generate_class(base_schema, ref_namespace, write_file=True, explicit_name=base_name)
        
        # Generate types for each choice
        choice_types = []
        choices = structure_schema.get('choices', {})
        
        for choice_key, choice_schema in choices.items():
            if isinstance(choice_schema, dict):
                if '$ref' in choice_schema:
                    # Resolve reference and generate the type
                    ref_schema = self.resolve_ref(choice_schema['$ref'], self.schema_doc if isinstance(self.schema_doc, dict) else None)
                    if ref_schema:
                        ref_path = choice_schema['$ref'].split('/')
                        ref_name = ref_path[-1]
                        ref_namespace = '.'.join(ref_path[2:-1]) if len(ref_path) > 3 else parent_namespace
                        qualified_name = self.generate_class(ref_schema, ref_namespace, write_file=True, explicit_name=ref_name)
                        import_types.add(qualified_name)
                        choice_types.append(qualified_name.split('.')[-1])
                elif 'type' in choice_schema:
                    # Generate inline type
                    ts_type = self.convert_structure_type_to_typescript(choice_name, choice_key, choice_schema, schema_namespace, import_types)
                    choice_types.append(ts_type)
        
        # Return Union type
        if len(choice_types) == 0:
            return 'any'
        elif len(choice_types) == 1:
            return choice_types[0]
        else:
            return ' | '.join(choice_types)

    def generate_tuple(self, structure_schema: Dict, parent_namespace: str, 
                      write_file: bool = True, explicit_name: str = '') -> str:
        """ Generates a TypeScript tuple type from JSON Structure tuple type """
        tuple_name = pascal(explicit_name if explicit_name else structure_schema.get('name', 'Tuple'))
        namespace = self.concat_namespace(self.base_package, structure_schema.get('namespace', parent_namespace)).lower()
        typescript_qualified_name = self.typescript_fully_qualified_name_from_structure_type(parent_namespace, tuple_name)
        
        if typescript_qualified_name in self.generated_types:
            return typescript_qualified_name

        import_types: Set[str] = set()
        tuple_items = structure_schema.get('items', [])
        item_types = []
        for idx, item in enumerate(tuple_items):
            item_type = self.convert_structure_type_to_typescript(
                tuple_name, f'item{idx}', item, namespace, import_types)
            item_types.append(item_type)

        # TypeScript tuples are just arrays with fixed length and types
        tuple_type = f"[{', '.join(item_types)}]"
        
        # Generate type alias
        tuple_definition = f"export type {tuple_name} = {tuple_type};\n"

        if write_file:
            self.write_to_file(namespace, tuple_name, tuple_definition)
        self.generated_types[typescript_qualified_name] = 'tuple'
        return typescript_qualified_name

    def generate_test_value(self, field: Dict) -> str:
        """Generates a test value for a given field in TypeScript"""
        import random
        field_type = field['type_no_null']
        
        # Map TypeScript types to test values
        test_values = {
            'string': '"test_string"',
            'number': str(random.randint(1, 100)),
            'bigint': 'BigInt(123)',
            'boolean': str(random.choice(['true', 'false'])).lower(),
            'Date': 'new Date()',
            'any': '{ test: "data" }',
            'null': 'null'
        }
        
        # Handle arrays
        if field_type.endswith('[]'):
            inner_type = field_type[:-2]
            if inner_type in test_values:
                return f"[{test_values[inner_type]}]"
            else:
                # For custom types, create empty array
                return f"[]"
        
        # Handle Set
        if field_type.startswith('Set<'):
            inner_type = field_type[4:-1]
            if inner_type in test_values:
                return f"new Set([{test_values[inner_type]}])"
            else:
                return f"new Set()"
        
        # Handle maps (objects with string index signature)
        if field_type.startswith('{ [key: string]:'):
            return '{}'
        
        # Return test value or construct object for custom types
        return test_values.get(field_type, f'{{}} as {field_type}')

    def generate_test_class(self, namespace: str, class_name: str, fields: List[Dict[str, Any]]) -> None:
        """Generates a unit test class for a TypeScript class"""
        # Get only required fields for the test
        required_fields = [f for f in fields if f['is_required']]
        
        # Generate test values for required fields
        for field in required_fields:
            field['test_value'] = self.generate_test_value(field)
        
        # Determine relative path from test directory to src
        namespace_path = namespace.replace('.', '/') if namespace else ''
        if namespace_path:
            relative_path = f"{namespace_path}/{class_name}"
        else:
            relative_path = class_name
        
        test_class_definition = process_template(
            "structuretots/test_class.ts.jinja",
            class_name=class_name,
            required_fields=required_fields,
            relative_path=relative_path
        )
        
        # Write test file
        test_dir = os.path.join(self.output_dir, "test")
        os.makedirs(test_dir, exist_ok=True)
        
        test_file_path = os.path.join(test_dir, f"{class_name}.test.ts")
        with open(test_file_path, 'w', encoding='utf-8') as f:
            f.write(test_class_definition)

    def write_to_file(self, namespace: str, type_name: str, content: str) -> None:
        """ Writes generated content to a TypeScript file """
        namespace_path = namespace.replace('.', '/')
        file_dir = os.path.join(self.output_dir, 'src', namespace_path)
        os.makedirs(file_dir, exist_ok=True)
        
        file_path = os.path.join(file_dir, f'{type_name}.ts')
        with open(file_path, 'w', encoding='utf-8') as f:
            f.write(content)

    def generate_package_json(self, package_name: str) -> None:
        """ Generates package.json file """
        package_json = process_template(
            "structuretots/package.json.jinja",
            package_name=package_name or 'generated-types',
        )
        
        with open(os.path.join(self.output_dir, 'package.json'), 'w', encoding='utf-8') as f:
            f.write(package_json)

    def generate_tsconfig(self) -> None:
        """ Generates tsconfig.json file """
        tsconfig = process_template("structuretots/tsconfig.json.jinja")
        with open(os.path.join(self.output_dir, 'tsconfig.json'), 'w', encoding='utf-8') as f:
            f.write(tsconfig)

    def generate_gitignore(self) -> None:
        """ Generates .gitignore file """
        gitignore = process_template("structuretots/gitignore.jinja")
        with open(os.path.join(self.output_dir, '.gitignore'), 'w', encoding='utf-8') as f:
            f.write(gitignore)

    def generate_index(self) -> None:
        """ Generates index.ts that exports all generated types """
        exports = []
        for qualified_name, type_kind in self.generated_types.items():
            type_name = qualified_name.split('.')[-1]
            namespace = '.'.join(qualified_name.split('.')[:-1])
            if namespace:
                # Lowercase the namespace to match the directory structure created by write_to_file
                relative_path = namespace.lower().replace('.', '/') + '/' + type_name
            else:
                relative_path = type_name
            exports.append(f"export * from './{relative_path}.js';")
        
        index_content = '\n'.join(exports) + '\n' if exports else ''
        
        src_dir = os.path.join(self.output_dir, 'src')
        os.makedirs(src_dir, exist_ok=True)
        with open(os.path.join(src_dir, 'index.ts'), 'w', encoding='utf-8') as f:
            f.write(index_content)

    def convert(self, structure_schema_path: str, output_dir: str, package_name: str = '') -> None:
        """ Converts a JSON Structure schema file to TypeScript classes """
        self.output_dir = output_dir
        
        # Load schema
        with open(structure_schema_path, 'r', encoding='utf-8') as f:
            schema = json.load(f)
        
        self.convert_schema(schema, output_dir, package_name)

    def convert_schema(self, schema: JsonNode, output_dir: str, package_name: str = '') -> None:
        """ Converts a JSON Structure schema to TypeScript classes """
        self.output_dir = output_dir
        self.schema_doc = schema
        
        # Register schema IDs
        self.register_schema_ids(self.schema_doc)
        
        # Process definitions
        if 'definitions' in self.schema_doc:
            for def_name, def_schema in self.schema_doc['definitions'].items():
                if isinstance(def_schema, dict):
                    self.generate_class_or_choice(def_schema, '', write_file=True, explicit_name=def_name)
        
        # Process root schema if it's an object or choice
        if 'type' in self.schema_doc:
            root_namespace = self.schema_doc.get('namespace', '')
            self.generate_class_or_choice(self.schema_doc, root_namespace, write_file=True)
        
        # Generate project files
        self.generate_package_json(package_name)
        self.generate_tsconfig()
        self.generate_gitignore()
        self.generate_index()


def convert_structure_to_typescript(structure_schema_path: str, ts_file_path: str, 
                                   package_name: str = '', typedjson_annotation: bool = False, 
                                   avro_annotation: bool = False) -> None:
    """
    Converts a JSON Structure schema to TypeScript classes.
    
    Args:
        structure_schema_path: Path to the JSON Structure schema file
        ts_file_path: Output directory for TypeScript files
        package_name: Package name for the generated TypeScript project
        typedjson_annotation: Whether to include TypedJSON annotations
        avro_annotation: Whether to include Avro annotations
    """
    converter = StructureToTypeScript(package_name, typedjson_annotation, avro_annotation)
    converter.convert(structure_schema_path, ts_file_path, package_name)


def convert_structure_schema_to_typescript(structure_schema: JsonNode, output_dir: str, 
                                          package_name: str = '', typedjson_annotation: bool = False, 
                                          avro_annotation: bool = False) -> None:
    """
    Converts a JSON Structure schema to TypeScript classes.
    
    Args:
        structure_schema: JSON Structure schema to convert
        output_dir: Output directory for TypeScript files
        package_name: Package name for the generated TypeScript project
        typedjson_annotation: Whether to include TypedJSON annotations
        avro_annotation: Whether to include Avro annotations
    """
    converter = StructureToTypeScript(package_name, typedjson_annotation, avro_annotation)
    converter.convert_schema(structure_schema, output_dir, package_name)



""" 
Convert Avro schema to SQL schema for various databases.
"""

import json
import sys
import os
from typing import Dict, List, cast
from avrotize.common import altname, camel

JsonNode = Dict[str,
                'JsonNode'] | List['JsonNode'] | str | bool | int | float | None


def convert_avro_to_sql(avro_schema_path, dbscript_file_path, db_dialect, emit_cloudevents_columns=False, schema_name: str = ''):
    """
    Converts an Avro schema to database schema for the specified DB dialect.

    Args:
        avro_schema_path (str): Path to the Avro schema file.
        dbscript_file_path (str): Path to the output SQL file.
        db_dialect (str): SQL/DB dialect. Supported: 'sqlserver', 'postgres', 'mysql', 'mariadb', 
                          'sqlite', 'oracle', 'db2', 'sqlanywhere', 'bigquery', 'snowflake', 
                          'redshift', 'cassandra', 'mongodb', 'dynamodb', 'elasticsearch', 
                          'couchdb', 'neo4j', 'firebase', 'cosmosdb', 'hbase'.
        emit_cloudevents_columns (bool): Whether to include cloud events columns.
        schema_name (str): Schema name (optional).

    Raises:
        ValueError: If the SQL dialect is unsupported.
    """
    if db_dialect not in ["sqlserver", "postgres", "mysql", "mariadb", "sqlite", "oracle", "db2",
                          "sqlanywhere", "bigquery", "snowflake", "redshift", "cassandra", "mongodb",
                          "dynamodb", "elasticsearch", "couchdb", "neo4j", "firebase", "cosmosdb", "hbase"]:
        print(f"Unsupported SQL dialect: {db_dialect}")
        sys.exit(1)

    schema_file = avro_schema_path
    if not schema_file:
        print("Please specify the avro schema file")
        sys.exit(1)
    with open(schema_file, "r", encoding="utf-8") as f:
        schema_json = f.read()

    schema_list = schema = json.loads(schema_json)

    if isinstance(schema, list):
        tables_sql = []
        for schema in schema_list:
            if not isinstance(schema, dict) or "type" not in schema or schema["type"] != "record":
                continue
            tables_sql.extend(generate_sql(
                schema, db_dialect, emit_cloudevents_columns, schema_list, schema_name))
        with open(dbscript_file_path, "w", encoding="utf-8") as sql_file:
            sql_file.write("\n".join(tables_sql))
    else:
        if not isinstance(schema, dict) or "type" not in schema or schema["type"] != "record":
            raise ValueError("Invalid Avro record schema")
        tables_sql = generate_sql(
            schema, db_dialect, emit_cloudevents_columns, schema_list, schema_name)
        with open(dbscript_file_path, "w", encoding="utf-8") as sql_file:
            sql_file.write("\n".join(tables_sql))


def generate_sql(
        schema: Dict[str, JsonNode],
        sql_dialect: str,
        emit_cloudevents_columns: bool,
        schema_list: List[Dict[str, JsonNode]],
        schema_name: str = '') -> List[str]:
    """
    Generates SQL schema statements for the given Avro schema.

    Args:
        schema (dict): Avro schema.
        sql_dialect (str): SQL dialect.
        emit_cloudevents_columns (bool): Whether to include cloud events columns.
        schema_list (list): List of all schemas.
        schema_name (str): Schema name (optional).

    Returns:
        list: List of SQL statements.
    """
    if sql_dialect in ["sqlserver", "postgres", "mysql", "mariadb", "sqlite", "oracle", "db2", "sqlanywhere",
                       "bigquery", "snowflake", "redshift"]:
        return generate_relational_sql(schema, sql_dialect, emit_cloudevents_columns, schema_list, schema_name)
    elif sql_dialect == "cassandra":
        return generate_cassandra_schema(schema, emit_cloudevents_columns, schema_name)
    else:
        raise ValueError(f"Unsupported SQL dialect: {sql_dialect}")


def generate_relational_sql(
        schema: Dict[str, JsonNode],
        sql_dialect: str,
        emit_cloudevents_columns: bool,
        schema_list: List[Dict[str, JsonNode]],
        schema_name: str = '') -> List[str]:
    """
    Generates relational SQL schema statements for the given Avro schema.

    Args:
        schema (dict): Avro schema.
        sql_dialect (str): SQL dialect.
        emit_cloudevents_columns (bool): Whether to include cloud events columns.
        schema_list (list): List of all schemas.
        schema_name (str): Schema name (optional).

    Returns:
        list: List of SQL statements.
    """
    namespace = str(schema.get("namespace", "")).replace('.', '_')
    plain_table_name = altname(
        schema, 'sql') or f"{namespace}_{schema['name']}"
    table_name = escape_name(plain_table_name, sql_dialect)
    fields: List[Dict[str, JsonNode]] = cast(
        List[Dict[str, JsonNode]], schema["fields"])
    unique_record_keys: List[str] = cast(List[str], schema.get("unique", []))

    table_comments = generate_table_comments_json(schema)
    column_comments = generate_column_comments_json(fields, schema_list)

    sql = []
    sql.append(f"CREATE TABLE {table_name} (")
    for field in fields:
        column_name = escape_name(
            altname(field, 'sql') or field["name"], sql_dialect)
        column_type = avro_type_to_sql_type(field["type"], sql_dialect)
        column_definition = f"{column_name} {column_type}"
        if field.get("unique", False):
            column_definition += f" {unique_clause(sql_dialect)}"
        if sql_dialect == "mysql" and field["name"] in column_comments:
            cmt = column_comments[str(field['name'])].replace("'", "''")
            column_definition += f" COMMENT '{cmt}'"
        sql.append(f"    {column_definition},")

    if emit_cloudevents_columns:
        sql.extend([
            f"    {escape_name('___type', sql_dialect)} {avro_type_to_sql_type('string', sql_dialect)} NOT NULL,",
            f"    {escape_name('___source', sql_dialect)} {avro_type_to_sql_type('string', sql_dialect)} NOT NULL,",
            f"    {escape_name('___id', sql_dialect)} {avro_type_to_sql_type('string', sql_dialect)} NOT NULL,",
            f"    {escape_name('___time', sql_dialect)} {avro_type_to_sql_type('timestamp', sql_dialect)} NULL,",
            f"    {escape_name('___subject', sql_dialect)} {avro_type_to_sql_type('string', sql_dialect)} NULL,"
        ])

    if unique_record_keys:
        unique_column_altnames = []
        if sql_dialect in ["mysql", "mariadb"]:
            for field in fields:
                if field["name"] in unique_record_keys:
                    column_type = avro_type_to_sql_type(
                        field["type"], sql_dialect)
                    if column_type in ["BLOB", "TEXT"]:
                        unique_column_altnames.append(escape_name(
                            altname(field, 'sql') + "(20)", sql_dialect))
                    else:
                        unique_column_altnames.append(
                            escape_name(altname(field, 'sql'), sql_dialect))
        else:
            unique_column_altnames = [escape_name(altname(
                field, 'sql'), sql_dialect) for field in fields if field["name"] in unique_record_keys]

        sql.append(
            f"    {primary_key_clause(sql_dialect)} ({', '.join(unique_column_altnames)})")
    else:
        sql[-1] = sql[-1][:-1]  # Remove the last comma
    sql.append(");")
    sql.append("")

    if sql_dialect != "mysql":
        sql.extend(generate_table_comment_sql(
            sql_dialect, table_comments, plain_table_name))
        sql.extend(generate_column_comment_sql(
            sql_dialect, column_comments, plain_table_name))

    return sql


def generate_table_comments_json(schema: Dict[str, JsonNode]) -> Dict[str, str]:
    """
    Generates table-level comments as JSON.

    Args:
        schema (dict): Avro schema.

    Returns:
        dict: Table-level comments as JSON.
    """
    comments = {}
    if "doc" in schema:
        comments["doc"] = str(schema["doc"])
    return comments


def generate_column_comments_json(fields: List[Dict[str, JsonNode]], schema_list: List[Dict[str, JsonNode]]) -> Dict[str, str]:
    """
    Generates column-level comments as JSON.

    Args:
        fields (list): List of fields.
        schema_list (list): List of all schemas.

    Returns:
        dict: Column-level comments as JSON.
    """
    comments = {}
    for field in fields:
        column_comment = {}
        if "doc" in field:
            column_comment["doc"] = field["doc"]
        column_type = field["type"]
        if isinstance(column_type, list):
            column_type = [x for x in column_type if x != "null"]
            if len(column_type) > 1:
                column_comment["schema"] = {
                    "type": "union", "types": column_type}
                column_type = "union"
            else:
                column_type = column_type[0]
        if isinstance(column_type, dict) and column_type["type"] in ["array", "map", "record", "enum", "fixed"]:
            column_comment["schema"] = column_type
        elif isinstance(schema_list, list):
            column_schema = next(
                (x for x in schema_list if x["name"] == column_type), None)
            if column_schema:
                column_comment["schema"] = column_schema
        comments[str(field["name"])] = json.dumps(column_comment)
    return comments


def generate_table_comment_sql(dialect: str, table_comments: Dict[str, str], table_name: str) -> List[str]:
    """
    Generates SQL statements for table-level comments.

    Args:
        dialect (str): SQL dialect.
        table_comments (dict): Table-level comments as JSON.
        table_name (str): Table name.

    Returns:
        list: SQL statements for table-level comments.
    """
    comments = []
    if "doc" in table_comments:
        doc_string = table_comments["doc"].replace("'", "''")
        if dialect == "sqlserver":
            comments.append(
                f"EXEC sp_addextendedproperty 'MS_Description', '{doc_string}', 'SCHEMA', 'dbo', 'TABLE', '{table_name}';")
        elif dialect in ["postgres", "oracle"]:
            comments.append(
                f"COMMENT ON TABLE {escape_name(table_name, dialect)} IS '{doc_string}';")
        elif dialect == "sqlite":
            comments.append(
                f"-- COMMENT ON TABLE {escape_name(table_name, dialect)} IS '{doc_string}';")
        # Add more dialect-specific clauses if needed
    return comments


def generate_column_comment_sql(dialect: str, column_comments: Dict[str, str], table_name: str) -> List[str]:
    """
    Generates SQL statements for column-level comments.

    Args:
        dialect (str): SQL dialect.
        column_comments (dict): Column-level comments as JSON.
        table_name (str): Table name.

    Returns:
        list: SQL statements for column-level comments.
    """
    comments = []
    for column_name, comment in column_comments.items():
        comment_data = json.loads(comment)
        doc = comment_data.get("doc", "")
        doc = doc.replace("'", "''")
        schema = comment_data.get("schema", "")
        if dialect == "sqlserver":
            if doc:
                comments.append(
                    f"EXEC sp_addextendedproperty 'MS_Description', '{doc}', 'SCHEMA', 'dbo', 'TABLE', '{table_name}', 'COLUMN', '{column_name}';")
            if schema:
                comments.append(
                    f"EXEC sp_addextendedproperty 'MS_Schema', '{json.dumps(schema)}', 'SCHEMA', 'dbo', 'TABLE', '{table_name}', 'COLUMN', '{column_name}';")
        else:
            comment = comment.replace("'", "''")
            comments.append(
                f"COMMENT ON COLUMN {escape_name(table_name, dialect)}.{escape_name(column_name, dialect)} IS '{comment}';")
    return comments


def escape_name(name, dialect):
    """
    Escapes a name (table or column) for the given SQL dialect.

    Args:
        name (str): The name to escape.
        dialect (str): The SQL dialect.

    Returns:
        str: The escaped name.
    """
    if dialect in ["sqlserver", "sqlanywhere"]:
        return f"[{name}]"
    elif dialect in ["postgres", "sqlite", "bigquery", "snowflake", "redshift"]:
        return f'"{name}"'
    elif dialect in ["mysql", "mariadb"]:
        return f"`{name}`"
    elif dialect in ["oracle", "db2"]:
        return f'"{name.upper()}"'
    elif dialect == "cassandra":
        return f'"{name}"'
    elif dialect in ["mongodb", "dynamodb", "elasticsearch", "couchdb", "neo4j", "firebase", "cosmosdb", "hbase"]:
        return name
    else:
        return name


def avro_type_to_sql_type(avro_type, dialect):
    """
    Maps an Avro type to a SQL type for the specified dialect.

    Args:
        avro_type (str or dict): The Avro type.
        dialect (str): The SQL dialect.

    Returns:
        str: The corresponding SQL type.
    """
    avro_to_sql_type_map = {
        "sqlserver": {
            "null": "NULL",
            "boolean": "BIT",
            "int": "INT",
            "long": "BIGINT",
            "float": "FLOAT",
            "double": "FLOAT",
            "bytes": "VARBINARY(MAX)",
            "string": "NVARCHAR(512)",
            "array": "NVARCHAR(MAX)",
            "map": "NVARCHAR(MAX)",
            "record": "NVARCHAR(MAX)",
            "union": "NVARCHAR(MAX)"
        },
        "postgres": {
            "null": "NULL",
            "boolean": "BOOLEAN",
            "int": "INTEGER",
            "long": "BIGINT",
            "float": "REAL",
            "double": "DOUBLE PRECISION",
            "bytes": "BYTEA",
            "string": "VARCHAR(512)",
            "array": "JSONB",
            "map": "JSONB",
            "record": "JSONB",
            "union": "JSONB"
        },
        "mysql": {
            "null": "NULL",
            "boolean": "BOOLEAN",
            "int": "INT",
            "long": "BIGINT",
            "float": "FLOAT",
            "double": "DOUBLE",
            "bytes": "BLOB",
            "string": "VARCHAR(512)",
            "array": "JSON",
            "map": "JSON",
            "record": "JSON",
            "union": "JSON"
        },
        "mariadb": {
            "null": "NULL",
            "boolean": "BOOLEAN",
            "int": "INT",
            "long": "BIGINT",
            "float": "FLOAT",
            "double": "DOUBLE",
            "bytes": "BLOB",
            "string": "VARCHAR(512)",
            "array": "JSON",
            "map": "JSON",
            "record": "JSON",
            "union": "JSON"
        },
        "sqlite": {
            "null": "NULL",
            "boolean": "BOOLEAN",
            "int": "INTEGER",
            "long": "INTEGER",
            "float": "REAL",
            "double": "REAL",
            "bytes": "BLOB",
            "string": "VARCHAR(512)",
            "array": "TEXT",
            "map": "TEXT",
            "record": "TEXT",
            "union": "TEXT"
        },
        "oracle": {
            "null": "NULL",
            "boolean": "NUMBER(1)",
            "int": "NUMBER(10)",
            "long": "NUMBER(19)",
            "float": "FLOAT(126)",
            "double": "FLOAT(126)",
            "bytes": "BLOB",
            "string": "VARCHAR(512)",
            "array": "CLOB",
            "map": "CLOB",
            "record": "CLOB",
            "union": "CLOB"
        },
        "db2": {
            "null": "NULL",
            "boolean": "BOOLEAN",
            "int": "INTEGER",
            "long": "BIGINT",
            "float": "REAL",
            "double": "DOUBLE",
            "bytes": "BLOB",
            "string": "VARCHAR(512)",
            "array": "CLOB",
            "map": "CLOB",
            "record": "CLOB",
            "union": "CLOB"
        },
        "sqlanywhere": {
            "null": "NULL",
            "boolean": "BIT",
            "int": "INTEGER",
            "long": "BIGINT",
            "float": "FLOAT",
            "double": "FLOAT",
            "bytes": "LONG BINARY",
            "string": "VARCHAR(512)",
            "array": "LONG VARCHAR",
            "map": "LONG VARCHAR",
            "record": "LONG VARCHAR",
            "union": "LONG VARCHAR"
        },
        "bigquery": {
            "null": "NULL",
            "boolean": "BOOL",
            "int": "INT64",
            "long": "INT64",
            "float": "FLOAT64",
            "double": "FLOAT64",
            "bytes": "BYTES",
            "string": "STRING",
            "array": "STRING",
            "map": "STRING",
            "record": "STRING",
            "union": "STRING"
        },
        "snowflake": {
            "null": "NULL",
            "boolean": "BOOLEAN",
            "int": "NUMBER",
            "long": "NUMBER",
            "float": "FLOAT",
            "double": "FLOAT",
            "bytes": "BINARY",
            "string": "STRING",
            "array": "VARIANT",
            "map": "VARIANT",
            "record": "VARIANT",
            "union": "VARIANT"
        },
        "redshift": {
            "null": "NULL",
            "boolean": "BOOLEAN",
            "int": "INTEGER",
            "long": "BIGINT",
            "float": "REAL",
            "double": "DOUBLE PRECISION",
            "bytes": "VARBYTE",
            "string": "VARCHAR(256)",
            "array": "VARCHAR(65535)",
            "map": "VARCHAR(65535)",
            "record": "VARCHAR(65535)",
            "union": "VARCHAR(65535)"
        }
    }

    if isinstance(avro_type, list):
        avro_type = [x for x in avro_type if x != "null"]
        if len(avro_type) > 1:
            return avro_to_sql_type_map[dialect]["union"]
        avro_type = avro_type[0]

    if isinstance(avro_type, dict):
        avro_type = avro_type.get("type", "string")

    return avro_to_sql_type_map[dialect].get(avro_type, avro_to_sql_type_map[dialect]["string"])


def unique_clause(dialect):
    """
    Returns the UNIQUE clause for the given SQL dialect.

    Args:
        dialect (str): The SQL dialect.

    Returns:
        str: The UNIQUE clause.
    """
    return "UNIQUE"


def primary_key_clause(dialect):
    """
    Returns the PRIMARY KEY clause for the given SQL dialect.

    Args:
        dialect (str): The SQL dialect.

    Returns:
        str: The PRIMARY KEY clause.
    """
    return "PRIMARY KEY"


def generate_cassandra_schema(schema: Dict[str, JsonNode], emit_cloudevents_columns: bool, schema_name: str) -> List[str]:
    """
    Generates Cassandra schema statements for the given Avro schema.

    Args:
        schema (dict): Avro schema.
        emit_cloudevents_columns (bool): Whether to include cloud events columns.
        schema_name (str): Schema name (optional).

    Returns:
        list: List of Cassandra schema statements.
    """
    namespace = cast(str, schema.get("namespace", "")).replace(".", "_")
    table_name = altname(schema, 'sql')
    table_name = escape_name(
        f"{namespace}_{table_name}" if namespace else table_name, "cassandra")
    table_name = f"{schema_name}.{table_name}" if schema_name else table_name
    # Cassandra table name length limit
    table_name = compact_table_name(table_name, 48)
    fields: List[Dict[str, JsonNode]] = cast(
        List[Dict[str, JsonNode]], schema["fields"])
    unique_record_keys: List[str] = cast(List[str], schema.get("unique", []))

    cql = []
    cql.append(f"CREATE TABLE {table_name} (")
    for field in fields:
        column_name = escape_name(
            altname(field, 'sql') or field["name"], 'cassandra')
        column_type = convert_avro_type_to_cassandra_type(field["type"])
        cql.append(f"    {column_name} {column_type},")
    if emit_cloudevents_columns:
        cql.extend([
            f"    {escape_name('cloudevents_type', 'cassandra')} text,",
            f"    {escape_name('cloudevents_source', 'cassandra')} text,",
            f"    {escape_name('cloudevents_id', 'cassandra')} text,",
            f"    {escape_name('cloudevents_time', 'cassandra')} timestamp,",
            f"    {escape_name('cloudevents_subject', 'cassandra')} text,"
        ])
    if unique_record_keys:
        unique_columns = [escape_name(field_name, "cassandra")
                          for field_name in unique_record_keys]
        unique_column_altnames = [altname(
            field, 'sql') for field in fields if field["name"] in unique_record_keys]
        cql.append(f"    PRIMARY KEY ({', '.join(unique_columns)})")
    elif emit_cloudevents_columns:
        cql.append(
            f"    PRIMARY KEY ({escape_name('cloudevents_id', 'cassandra')})")
    else:
        all_columns = [escape_name(
            altname(field, 'sql') or field, 'cassandra') for field in fields]
        cql.append(f"    PRIMARY KEY ({', '.join(all_columns)})")
    cql.append(");")
    return cql


def convert_avro_type_to_cassandra_type(avro_type):
    """
    Converts an Avro type to Cassandra type.

    Args:
        avro_type (str or dict): The Avro type.

    Returns:
        str: The corresponding Cassandra type.
    """
    avro_to_cassandra_type_map = {
        "null": "NULL",
        "boolean": "boolean",
        "int": "int",
        "long": "bigint",
        "float": "float",
        "double": "double",
        "bytes": "blob",
        "string": "text",
        "array": "text",
        "map": "text",
        "record": "text",
        "union": "text"
    }

    if isinstance(avro_type, list):
        avro_type = [x for x in avro_type if x != "null"]
        if len(avro_type) > 1:
            return avro_to_cassandra_type_map["union"]
        avro_type = avro_type[0]

    if isinstance(avro_type, dict):
        avro_type = avro_type.get("type", "string")

    return avro_to_cassandra_type_map.get(avro_type, "text")


def compact_table_name(table_name, max_length):
    """
    Compacts the table name to fit the specified maximum length.

    Args:
        table_name (str): Table name.
        max_length (int): Maximum length of the table name.

    Returns:
        str: Compacted table name.
    """
    if len(table_name) > max_length:
        # Drop vowels, one by one, seek from end
        vowels = "aeiou"
        while len(table_name) > max_length:
            count = 0
            for i in range(len(table_name) - 1, -1, -1):
                if table_name[i].lower() in vowels:
                    count += 1
                    table_name = table_name[:i] + table_name[i + 1:]
                    break
            if count == 0:
                break
    return table_name[:max_length]


def convert_avro_to_nosql(avro_schema_path, nosql_file_path, nosql_dialect, emit_cloudevents_columns=False):
    """
    Converts an Avro schema to NoSQL schema for the specified NoSQL dialect.

    Args:
        avro_schema_path (str): Path to the Avro schema file.
        nosql_file_path (str): Path to the output NoSQL schema file.
        nosql_dialect (str): NoSQL dialect (e.g., 'mongodb', 'dynamodb', 'cassandra').
        emit_cloudevents_columns (bool): Whether to include cloud events columns.

    Raises:
        ValueError: If the NoSQL dialect is unsupported.
    """
    schema_file = avro_schema_path
    if not schema_file:
        print("Please specify the avro schema file")
        sys.exit(1)
    with open(schema_file, "r", encoding="utf-8") as f:
        schema_json = f.read()

    schema_list = schema = json.loads(schema_json)
    dirname = nosql_file_path
    if not os.path.exists(dirname):
        os.makedirs(dirname, exist_ok=True)

    if isinstance(schema, list):
        for schema in schema_list:
            if not isinstance(schema, dict) or "type" not in schema or schema["type"] != "record":
                continue
            model = generate_nosql(schema, nosql_dialect,
                                   emit_cloudevents_columns, schema_list)
            file_name = os.path.join(
                nosql_file_path, get_file_name(schema, get_nosql_file_extension(nosql_dialect)))
            with open(file_name, "w", encoding="utf-8") as nosql_file:
                if isinstance(model, list):
                    nosql_file.write("\n".join(model))
                else:
                    nosql_file.write(model)
    else:
        if not isinstance(schema, dict) or "type" not in schema or schema["type"] != "record":
            raise ValueError("Invalid Avro record schema")
        model = generate_nosql(schema, nosql_dialect,
                               emit_cloudevents_columns, schema_list)
        file_name = os.path.join(
            nosql_file_path, get_file_name(schema_list, get_nosql_file_extension(nosql_dialect)))
        with open(file_name, "w", encoding="utf-8") as nosql_file:
            nosql_file.write(model)

def get_nosql_file_extension(nosql_dialect):
    """
    Returns the file extension for the given NoSQL dialect.

    Args:
        nosql_dialect (str): NoSQL dialect.

    Returns:
        str: File extension.
    """
    if nosql_dialect == "neo4j":
        return "cypher"
    else:
        return "json"


def generate_nosql(schema, nosql_dialect, emit_cloudevents_columns, schema_list):
    """
    Generates NoSQL schema statements for the given Avro schema.

    Args:
        schema (dict): Avro schema.
        nosql_dialect (str): NoSQL dialect.
        emit_cloudevents_columns (bool): Whether to include cloud events columns.
        schema_list (list): List of all schemas.

    Returns:
        list: List of NoSQL schema statements.
    """
    if nosql_dialect == "mongodb":
        return generate_mongodb_schema(schema, emit_cloudevents_columns)
    elif nosql_dialect == "dynamodb":
        return generate_dynamodb_schema(schema, emit_cloudevents_columns)
    elif nosql_dialect == "elasticsearch":
        return generate_elasticsearch_schema(schema, emit_cloudevents_columns)
    elif nosql_dialect == "couchdb":
        return generate_couchdb_schema(schema, emit_cloudevents_columns)
    elif nosql_dialect == "neo4j":
        return generate_neo4j_schema(schema, emit_cloudevents_columns)
    elif nosql_dialect == "firebase":
        return generate_firebase_schema(schema, emit_cloudevents_columns)
    elif nosql_dialect == "cosmosdb":
        return generate_cosmosdb_schema(schema, emit_cloudevents_columns)
    elif nosql_dialect == "hbase":
        return generate_hbase_schema(schema, emit_cloudevents_columns)
    else:
        raise ValueError(f"Unsupported NoSQL dialect: {nosql_dialect}")


def generate_mongodb_schema(schema, emit_cloudevents_columns):
    """
    Generates MongoDB schema statements for the given Avro schema.

    Args:
        schema (dict): Avro schema.
        emit_cloudevents_columns (bool): Whether to include cloud events columns.

    Returns:
        list: List of MongoDB schema statements.
    """
    namespace = schema.get("namespace", "")
    collection_name = altname(schema, 'sql') or f"{namespace}_{schema['name']}"
    collection_name = namespace + "." + \
        collection_name if namespace else collection_name
    collection_name = '.'.join(camel(t) for t in collection_name.split('.'))
    fields = schema["fields"]
    unique_record_keys = schema.get("unique", [])

    mongodb_schema = {
        "$jsonSchema": {
            "bsonType": "object",
            "required": [],
            "properties": {}
        }
    }

    for field in fields:
        column_name = altname(field, 'sql') or field["name"]
        column_type = convert_avro_type_to_mongodb_type(field["type"])
        mongodb_schema["$jsonSchema"]["properties"][column_name] = column_type
        if "null" not in field["type"]:
            mongodb_schema["$jsonSchema"]["required"].append(column_name)
        if field.get("unique", False):
            mongodb_schema["$jsonSchema"]["properties"][column_name]["unique"] = True
        if field.get("doc", ""):
            mongodb_schema["$jsonSchema"]["properties"][column_name]["description"] = field["doc"]

    if emit_cloudevents_columns:
        mongodb_schema["$jsonSchema"]["properties"].update({
            "___type": {"bsonType": "string"},
            "___source": {"bsonType": "string"},
            "___id": {"bsonType": "string"},
            "___time": {"bsonType": "date"},
            "___subject": {"bsonType": "string"}
        })
        mongodb_schema["$jsonSchema"]["required"].extend(
            ["___type", "___source", "___id"])

    return json.dumps({collection_name: mongodb_schema}, indent=4)


def convert_avro_type_to_mongodb_type(avro_type):
    """
    Converts an Avro type to MongoDB type.

    Args:
        avro_type (str or dict): The Avro type.

    Returns:
        dict: The corresponding MongoDB type.
    """
    avro_to_mongodb_type_map = {
        "null": {"bsonType": "null"},
        "boolean": {"bsonType": "bool"},
        "int": {"bsonType": "int"},
        "long": {"bsonType": "long"},
        "float": {"bsonType": "double"},
        "double": {"bsonType": "double"},
        "bytes": {"bsonType": "binData"},
        "string": {"bsonType": "string"},
        "array": {"bsonType": "array"},
        "map": {"bsonType": "object"},
        "record": {"bsonType": "object"},
        "union": {"bsonType": "object"}
    }

    if isinstance(avro_type, list):
        avro_type = [x for x in avro_type if x != "null"]
        if len(avro_type) > 1:
            return avro_to_mongodb_type_map["union"]
        avro_type = avro_type[0]

    if isinstance(avro_type, dict):
        avro_type = avro_type.get("type", "string")

    return avro_to_mongodb_type_map.get(avro_type, {"bsonType": "string"})


def generate_dynamodb_schema(schema, emit_cloudevents_columns):
    """
    Generates DynamoDB schema statements for the given Avro schema.

    Args:
        schema (dict): Avro schema.
        emit_cloudevents_columns (bool): Whether to include cloud events columns.

    Returns:
        list: List of DynamoDB schema statements.
    """
    namespace = schema.get("namespace", "").replace('.', '_')
    table_name = altname(schema, 'sql') or f"{namespace}_{schema['name']}"
    fields = schema["fields"]
    unique_record_keys = schema.get("unique", [])

    dynamodb_schema = {
        "TableName": table_name,
        "KeySchema": [],
        "AttributeDefinitions": [],
        "ProvisionedThroughput": {
            "ReadCapacityUnits": 5,
            "WriteCapacityUnits": 5
        }
    }

    for field in fields:
        column_name = altname(field, 'sql') or field["name"]
        column_type = convert_avro_type_to_dynamodb_type(field["type"])
        dynamodb_schema["AttributeDefinitions"].append(
            {"AttributeName": column_name, "AttributeType": column_type})
        if not dynamodb_schema["KeySchema"]:
            dynamodb_schema["KeySchema"].append(
                {"AttributeName": column_name, "KeyType": "HASH"})
        if field.get("unique", False):
            dynamodb_schema["AttributeDefinitions"].append(
                {"AttributeName": column_name, "AttributeType": column_type})
            dynamodb_schema["KeySchema"].append(
                {"AttributeName": column_name, "KeyType": "RANGE"})

    if emit_cloudevents_columns:
        dynamodb_schema["AttributeDefinitions"].extend([
            {"AttributeName": "___type", "AttributeType": "S"},
            {"AttributeName": "___source", "AttributeType": "S"},
            {"AttributeName": "___id", "AttributeType": "S"},
            {"AttributeName": "___time", "AttributeType": "S"},
            {"AttributeName": "___subject", "AttributeType": "S"}
        ])
        dynamodb_schema["KeySchema"].append(
            {"AttributeName": "___id", "KeyType": "HASH"})

    return json.dumps(dynamodb_schema, indent=4)


def convert_avro_type_to_dynamodb_type(avro_type):
    """
    Converts an Avro type to DynamoDB type.

    Args:
        avro_type (str or dict): The Avro type.

    Returns:
        str: The corresponding DynamoDB type.
    """
    avro_to_dynamodb_type_map = {
        "null": "NULL",
        "boolean": "BOOL",
        "int": "N",
        "long": "N",
        "float": "N",
        "double": "N",
        "bytes": "B",
        "string": "S",
        "array": "S",
        "map": "S",
        "record": "S",
        "union": "S"
    }

    if isinstance(avro_type, list):
        avro_type = [x for x in avro_type if x != "null"]
        if len(avro_type) > 1:
            return avro_to_dynamodb_type_map["union"]
        avro_type = avro_type[0]

    if isinstance(avro_type, dict):
        avro_type = avro_type.get("type", "string")

    return avro_to_dynamodb_type_map.get(avro_type, "S")


def generate_elasticsearch_schema(schema, emit_cloudevents_columns):
    """
    Generates Elasticsearch schema statements for the given Avro schema.

    Args:
        schema (dict): Avro schema.
        emit_cloudevents_columns (bool): Whether to include cloud events columns.

    Returns:
        list: List of Elasticsearch schema statements.
    """
    namespace = schema.get("namespace", "").replace('.', '_')
    index_name = altname(schema, 'sql') or f"{namespace}_{schema['name']}"
    fields = schema["fields"]

    es_mapping = {
        "mappings": {
            "properties": {}
        }
    }

    for field in fields:
        column_name = altname(field, 'sql') or field["name"]
        column_type = convert_avro_type_to_elasticsearch_type(field["type"])
        es_mapping["mappings"]["properties"][column_name] = column_type

    if emit_cloudevents_columns:
        es_mapping["mappings"]["properties"].update({
            "___type": {"type": "keyword"},
            "___source": {"type": "keyword"},
            "___id": {"type": "keyword"},
            "___time": {"type": "date"},
            "___subject": {"type": "keyword"}
        })

    return json.dumps({index_name: es_mapping}, indent=4)


def convert_avro_type_to_elasticsearch_type(avro_type):
    """
    Converts an Avro type to Elasticsearch type.

    Args:
        avro_type (str or dict): The Avro type.

    Returns:
        dict: The corresponding Elasticsearch type.
    """
    avro_to_elasticsearch_type_map = {
        "null": {"type": "null"},
        "boolean": {"type": "boolean"},
        "int": {"type": "integer"},
        "long": {"type": "long"},
        "float": {"type": "float"},
        "double": {"type": "double"},
        "bytes": {"type": "binary"},
        "string": {"type": "text"},
        "array": {"type": "text"},
        "map": {"type": "text"},
        "record": {"type": "text"},
        "union": {"type": "text"}
    }

    if isinstance(avro_type, list):
        avro_type = [x for x in avro_type if x != "null"]
        if len(avro_type) > 1:
            return avro_to_elasticsearch_type_map["union"]
        avro_type = avro_type[0]

    if isinstance(avro_type, dict):
        avro_type = avro_type.get("type", "string")

    return avro_to_elasticsearch_type_map.get(avro_type, {"type": "text"})


def generate_couchdb_schema(schema, emit_cloudevents_columns):
    """
    Generates CouchDB schema statements for the given Avro schema.

    Args:
        schema (dict): Avro schema.
        emit_cloudevents_columns (bool): Whether to include cloud events columns.

    Returns:
        list: List of CouchDB schema statements.
    """
    namespace = schema.get("namespace", "").replace('.', '_')
    db_name = altname(schema, 'sql') or f"{namespace}_{schema['name']}"
    fields = schema["fields"]

    couchdb_schema = {
        "type": "object",
        "properties": {}
    }

    for field in fields:
        column_name = altname(field, 'sql') or field["name"]
        column_type = convert_avro_type_to_couchdb_type(field["type"])
        couchdb_schema["properties"][column_name] = column_type

    if emit_cloudevents_columns:
        couchdb_schema["properties"].update({
            "___type": {"type": "string"},
            "___source": {"type": "string"},
            "___id": {"type": "string"},
            "___time": {"type": "string"},
            "___subject": {"type": "string"}
        })

    return json.dumps({db_name: couchdb_schema}, indent=4)


def convert_avro_type_to_couchdb_type(avro_type):
    """
    Converts an Avro type to CouchDB type.

    Args:
        avro_type (str or dict): The Avro type.

    Returns:
        dict: The corresponding CouchDB type.
    """
    avro_to_couchdb_type_map = {
        "null": {"type": "null"},
        "boolean": {"type": "boolean"},
        "int": {"type": "integer"},
        "long": {"type": "integer"},
        "float": {"type": "number"},
        "double": {"type": "number"},
        "bytes": {"type": "string"},
        "string": {"type": "string"},
        "array": {"type": "string"},
        "map": {"type": "string"},
        "record": {"type": "string"},
        "union": {"type": "string"}
    }

    if isinstance(avro_type, list):
        avro_type = [x for x in avro_type if x != "null"]
        if len(avro_type) > 1:
            return avro_to_couchdb_type_map["union"]
        avro_type = avro_type[0]

    if isinstance(avro_type, dict):
        avro_type = avro_type.get("type", "string")

    return avro_to_couchdb_type_map.get(avro_type, {"type": "string"})


def generate_neo4j_schema(schema, emit_cloudevents_columns):
    """
    Generates Neo4j schema statements for the given Avro schema.

    Args:
        schema (dict): Avro schema.
        emit_cloudevents_columns (bool): Whether to include cloud events columns.

    Returns:
        list: List of Neo4j schema statements.
    """
    namespace = schema.get("namespace", "").replace('.', '_')
    label_name = altname(schema, 'sql') or f"{namespace}_{schema['name']}"
    fields = schema["fields"]

    cypher = []
    cypher.append(f"CREATE (:{label_name} {{")
    for field in fields:
        column_name = altname(field, 'sql') or field["name"]
        column_type = convert_avro_type_to_neo4j_type(field["type"])
        cypher.append(f"    {column_name}: {column_type},")
    if emit_cloudevents_columns:
        cypher.extend([
            f"    {escape_name('___type', 'neo4j')}: 'string',",
            f"    {escape_name('___source', 'neo4j')}: 'string',",
            f"    {escape_name('___id', 'neo4j')}: 'string',",
            f"    {escape_name('___time', 'neo4j')}: 'datetime',",
            f"    {escape_name('___subject', 'neo4j')}: 'string'"
        ])
    cypher[-1] = cypher[-1][:-1]  # Remove the last comma
    cypher.append("});")
    return cypher


def convert_avro_type_to_neo4j_type(avro_type):
    """
    Converts an Avro type to Neo4j type.

    Args:
        avro_type (str or dict): The Avro type.

    Returns:
        str: The corresponding Neo4j type.
    """
    avro_to_neo4j_type_map = {
        "null": "NULL",
        "boolean": "boolean",
        "int": "integer",
        "long": "long",
        "float": "float",
        "double": "float",
        "bytes": "string",
        "string": "string",
        "array": "string",
        "map": "string",
        "record": "string",
        "union": "string"
    }

    if isinstance(avro_type, list):
        avro_type = [x for x in avro_type if x != "null"]
        if len(avro_type) > 1:
            return avro_to_neo4j_type_map["union"]
        avro_type = avro_type[0]

    if isinstance(avro_type, dict):
        avro_type = avro_type.get("type", "string")

    return avro_to_neo4j_type_map.get(avro_type, "string")


def generate_firebase_schema(schema, emit_cloudevents_columns):
    """
    Generates Firebase schema statements for the given Avro schema.

    Args:
        schema (dict): Avro schema.
        emit_cloudevents_columns (bool): Whether to include cloud events columns.

    Returns:
        list: List of Firebase schema statements.
    """
    namespace = schema.get("namespace", "").replace('.', '_')
    collection_name = altname(schema, 'sql') or f"{namespace}_{schema['name']}"
    fields = schema["fields"]

    firebase_schema = {
        "fields": {}
    }

    for field in fields:
        column_name = altname(field, 'sql') or field["name"]
        column_type = convert_avro_type_to_firebase_type(field["type"])
        firebase_schema["fields"][column_name] = column_type

    if emit_cloudevents_columns:
        firebase_schema["fields"].update({
            "___type": {"type": "string"},
            "___source": {"type": "string"},
            "___id": {"type": "string"},
            "___time": {"type": "timestamp"},
            "___subject": {"type": "string"}
        })

    return json.dumps({collection_name: firebase_schema}, indent=4)


def convert_avro_type_to_firebase_type(avro_type):
    """
    Converts an Avro type to Firebase type.

    Args:
        avro_type (str or dict): The Avro type.

    Returns:
        dict: The corresponding Firebase type.
    """
    avro_to_firebase_type_map = {
        "null": {"type": "null"},
        "boolean": {"type": "boolean"},
        "int": {"type": "integer"},
        "long": {"type": "integer"},
        "float": {"type": "number"},
        "double": {"type": "number"},
        "bytes": {"type": "string"},
        "string": {"type": "string"},
        "array": {"type": "string"},
        "map": {"type": "string"},
        "record": {"type": "string"},
        "union": {"type": "string"}
    }

    if isinstance(avro_type, list):
        avro_type = [x for x in avro_type if x != "null"]
        if len(avro_type) > 1:
            return avro_to_firebase_type_map["union"]
        avro_type = avro_type[0]

    if isinstance(avro_type, dict):
        avro_type = avro_type.get("type", "string")

    return avro_to_firebase_type_map.get(avro_type, {"type": "string"})


def generate_cosmosdb_schema(schema, emit_cloudevents_columns):
    """
    Generates CosmosDB schema statements for the given Avro schema.

    Args:
        schema (dict): Avro schema.
        emit_cloudevents_columns (bool): Whether to include cloud events columns.

    Returns:
        list: List of CosmosDB schema statements.
    """
    namespace = schema.get("namespace", "").replace('.', '_')
    collection_name = altname(schema, 'sql') or f"{namespace}_{schema['name']}"
    fields = schema["fields"]

    cosmosdb_schema = {
        "id": collection_name,
        "partitionKey": {
            "paths": [],
            "kind": "Hash"
        },
        "uniqueKeyPolicy": {
            "uniqueKeys": []
        },
        "fields": {}
    }

    for field in fields:
        column_name = altname(field, 'sql') or field["name"]
        column_type = convert_avro_type_to_cosmosdb_type(field["type"])
        cosmosdb_schema["fields"][column_name] = column_type
        cosmosdb_schema["partitionKey"]["paths"].append(f"/{column_name}")

    if emit_cloudevents_columns:
        cosmosdb_schema["fields"].update({
            "___type": {"type": "string"},
            "___source": {"type": "string"},
            "___id": {"type": "string"},
            "___time": {"type": "string"},
            "___subject": {"type": "string"}
        })
        cosmosdb_schema["partitionKey"]["paths"].append("/___id")

    return json.dumps(cosmosdb_schema, indent=4)


def convert_avro_type_to_cosmosdb_type(avro_type):
    """
    Converts an Avro type to CosmosDB type.

    Args:
        avro_type (str or dict): The Avro type.

    Returns:
        dict: The corresponding CosmosDB type.
    """
    avro_to_cosmosdb_type_map = {
        "null": {"type": "null"},
        "boolean": {"type": "boolean"},
        "int": {"type": "number"},
        "long": {"type": "number"},
        "float": {"type": "number"},
        "double": {"type": "number"},
        "bytes": {"type": "string"},
        "string": {"type": "string"},
        "array": {"type": "string"},
        "map": {"type": "string"},
        "record": {"type": "string"},
        "union": {"type": "string"}
    }

    if isinstance(avro_type, list):
        avro_type = [x for x in avro_type if x != "null"]
        if len(avro_type) > 1:
            return avro_to_cosmosdb_type_map["union"]
        avro_type = avro_type[0]

    if isinstance(avro_type, dict):
        avro_type = avro_type.get("type", "string")

    return avro_to_cosmosdb_type_map.get(avro_type, {"type": "string"})


def generate_hbase_schema(schema, emit_cloudevents_columns):
    """
    Generates HBase schema statements for the given Avro schema.

    Args:
        schema (dict): Avro schema.
        emit_cloudevents_columns (bool): Whether to include cloud events columns.

    Returns:
        list: List of HBase schema statements.
    """
    namespace = schema.get("namespace", "").replace('.', '_')
    table_name = altname(schema, 'sql') or f"{namespace}_{schema['name']}"
    fields = schema["fields"]

    hbase_schema = {
        "table": table_name,
        "column_families": []
    }

    for field in fields:
        column_name = altname(field, 'sql') or field["name"]
        column_family = convert_avro_type_to_hbase_column_family(field["type"])
        hbase_schema["column_families"].append(
            {"name": column_name, "column_family": column_family})

    if emit_cloudevents_columns:
        hbase_schema["column_families"].extend([
            {"name": "___type", "column_family": "string"},
            {"name": "___source", "column_family": "string"},
            {"name": "___id", "column_family": "string"},
            {"name": "___time", "column_family": "string"},
            {"name": "___subject", "column_family": "string"}
        ])

    return json.dumps(hbase_schema, indent=4)


def convert_avro_type_to_hbase_column_family(avro_type):
    """
    Converts an Avro type to HBase column family.

    Args:
        avro_type (str or dict): The Avro type.

    Returns:
        str: The corresponding HBase column family.
    """
    avro_to_hbase_column_family_map = {
        "null": "string",
        "boolean": "boolean",
        "int": "integer",
        "long": "integer",
        "float": "number",
        "double": "number",
        "bytes": "string",
        "string": "string",
        "array": "string",
        "map": "string",
        "record": "string",
        "union": "string"
    }

    if isinstance(avro_type, list):
        avro_type = [x for x in avro_type if x != "null"]
        if len(avro_type) > 1:
            return avro_to_hbase_column_family_map["union"]
        avro_type = avro_type[0]

    if isinstance(avro_type, dict):
        avro_type = avro_type.get("type", "string")

    return avro_to_hbase_column_family_map.get(avro_type, "string")


def get_file_name(avro_schema: dict, extension: str) -> str:
    """
    Generates a file name based on the Avro schema.

    Args:
        avro_schema (dict): Avro schema.
        extension (str): File extension.

    Returns:
        str: Generated file name.
    """
    namespace = avro_schema.get("namespace", "").replace('.', '_')
    name = avro_schema.get("name", "")
    return (namespace + '.' + name + '.' + extension) if namespace else (name + '.' + extension)



"""
OpenAPI/Swagger to JSON Structure converter.

This module converts OpenAPI 3.x and Swagger 2.0 documents to JSON Structure format 
by extracting schema definitions and delegating to the existing JSON Schema to 
JSON Structure converter.

Supported versions:
- Swagger 2.0
- OpenAPI 3.0.x
- OpenAPI 3.1.x
"""

# pylint: disable=line-too-long

import json
import os
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse

import requests

from avrotize.jsonstostructure import JsonToStructureConverter


class OpenApiToStructureConverter:
    """
    Converts OpenAPI 3.x and Swagger 2.0 documents to JSON Structure format.
    
    This converter extracts schema definitions from API documents and converts 
    them to JSON Structure format using the existing JSON Schema to JSON Structure 
    conversion machinery.
    
    Supported document types:
    - Swagger 2.0: Extracts from `definitions`
    - OpenAPI 3.0.x: Extracts from `components.schemas`
    - OpenAPI 3.1.x: Extracts from `components.schemas` (JSON Schema compatible)
    
    Attributes:
        root_namespace: The namespace for the root schema.
        root_class_name: The name of the root class.
        preserve_composition: Flag to preserve composition keywords.
        detect_inheritance: Flag to detect inheritance patterns.
        detect_discriminators: Flag to detect OpenAPI discriminator patterns.
        convert_empty_objects_to_maps: Flag to convert objects with only additionalProperties to maps.
        lift_inline_schemas: Flag to lift inline schemas from paths to definitions.
    """
    
    # Supported specification versions
    SWAGGER_2 = 'swagger_2'
    OPENAPI_3_0 = 'openapi_3_0'
    OPENAPI_3_1 = 'openapi_3_1'
    
    def __init__(self) -> None:
        """Initialize the OpenAPI to JSON Structure converter."""
        self.root_namespace = 'example.com'
        self.root_class_name = 'document'
        self.preserve_composition = False  # Resolve composition for JSON Structure Core compliance
        self.detect_inheritance = True
        self.detect_discriminators = True
        self.convert_empty_objects_to_maps = True
        self.lift_inline_schemas = False  # Optional: lift inline schemas from paths
        self.content_cache: Dict[str, str] = {}
    
    def detect_spec_version(self, doc: dict) -> Tuple[str, str]:
        """
        Detect the specification version of an API document.
        
        Args:
            doc: The API document as a dictionary.
            
        Returns:
            Tuple of (spec_type, version_string) where spec_type is one of:
            SWAGGER_2, OPENAPI_3_0, OPENAPI_3_1
            
        Raises:
            ValueError: If the document is not a valid Swagger/OpenAPI document.
        """
        # Check for Swagger 2.0
        if 'swagger' in doc:
            swagger_version = str(doc['swagger'])
            if swagger_version.startswith('2.'):
                return (self.SWAGGER_2, swagger_version)
            raise ValueError(f"Unsupported Swagger version: {swagger_version}. Only Swagger 2.0 is supported.")
        
        # Check for OpenAPI 3.x
        if 'openapi' in doc:
            openapi_version = str(doc['openapi'])
            if openapi_version.startswith('3.0'):
                return (self.OPENAPI_3_0, openapi_version)
            if openapi_version.startswith('3.1'):
                return (self.OPENAPI_3_1, openapi_version)
            if openapi_version.startswith('3.'):
                # Future 3.x versions - treat as 3.1 compatible
                return (self.OPENAPI_3_1, openapi_version)
            raise ValueError(f"Unsupported OpenAPI version: {openapi_version}. Supported versions: 3.0.x, 3.1.x")
        
        raise ValueError("Not a valid Swagger/OpenAPI document: missing 'swagger' or 'openapi' version field")
    
    def fetch_content(self, url: str) -> str:
        """
        Fetch content from a URL or file path.
        
        Args:
            url: The URL or file path to fetch content from.
            
        Returns:
            The content as a string.
            
        Raises:
            requests.RequestException: If there is an error fetching from HTTP/HTTPS.
            FileNotFoundError: If the file does not exist.
        """
        if url in self.content_cache:
            return self.content_cache[url]
        
        parsed_url = urlparse(url)
        
        if parsed_url.scheme in ['http', 'https']:
            response = requests.get(url, timeout=30)
            response.raise_for_status()
            content = response.text
            self.content_cache[url] = content
            return content
        elif parsed_url.scheme == 'file' or not parsed_url.scheme:
            # Handle file URLs or local paths
            file_path = parsed_url.path if parsed_url.scheme == 'file' else url
            if os.name == 'nt' and file_path.startswith('/'):
                file_path = file_path[1:]
            with open(file_path, 'r', encoding='utf-8') as f:
                content = f.read()
                self.content_cache[url] = content
                return content
        else:
            raise ValueError(f"Unsupported URL scheme: {parsed_url.scheme}")
    
    def extract_schemas_from_swagger2(self, swagger_doc: dict) -> dict:
        """
        Extract schemas from a Swagger 2.0 document and convert to JSON Schema format.
        
        Swagger 2.0 stores definitions at the root level under `definitions`.
        
        Args:
            swagger_doc: The Swagger 2.0 document as a dictionary.
            
        Returns:
            A JSON Schema document with definitions from the Swagger document.
        """
        json_schema: Dict[str, Any] = {
            "$schema": "http://json-schema.org/draft-04/schema#"
        }
        
        # Copy Swagger info to JSON Schema metadata
        if 'info' in swagger_doc:
            info = swagger_doc['info']
            if 'title' in info:
                json_schema['title'] = info['title']
            if 'description' in info:
                json_schema['description'] = info['description']
            if 'version' in info:
                title_slug = info.get('title', 'swagger').lower().replace(' ', '-')
                json_schema['$id'] = f"https://{self.root_namespace}/schemas/{title_slug}/{info['version']}"
        
        # Swagger 2.0 uses top-level 'definitions' 
        definitions = swagger_doc.get('definitions', {})
        
        if definitions:
            json_schema['definitions'] = {}
            for schema_name, schema_def in definitions.items():
                # Process the schema to handle Swagger-specific keywords
                processed_schema = self._process_swagger2_schema(schema_def)
                json_schema['definitions'][schema_name] = processed_schema
        
        # Optionally lift inline schemas from paths
        if self.lift_inline_schemas:
            self._lift_inline_schemas_from_swagger2_paths(swagger_doc, json_schema)
        
        return json_schema
    
    def _process_swagger2_schema(self, schema: Union[dict, Any]) -> Union[dict, Any]:
        """
        Process a Swagger 2.0 schema to handle Swagger-specific keywords.
        
        Swagger 2.0 uses slightly different keywords than OpenAPI 3.x.
        Notable differences:
        - No 'nullable' keyword (use x-nullable extension)
        - Uses 'type: file' for file uploads
        - discriminator is a string (property name), not an object
        
        Args:
            schema: The Swagger 2.0 schema object.
            
        Returns:
            The processed schema compatible with JSON Schema.
        """
        if not isinstance(schema, dict):
            return schema
        
        processed: Dict[str, Any] = {}
        
        for key, value in schema.items():
            if key == 'x-nullable':
                # Swagger extension for nullable
                if value is True:
                    processed['nullable'] = True
            elif key == 'discriminator':
                # Swagger 2.0 discriminator is just a property name string
                if isinstance(value, str):
                    processed['discriminator'] = {
                        'propertyName': value
                    }
                else:
                    processed['discriminator'] = value
            elif key == 'readOnly':
                if 'x-metadata' not in processed:
                    processed['x-metadata'] = {}
                processed['x-metadata']['readOnly'] = value
            elif key == 'xml':
                if 'x-metadata' not in processed:
                    processed['x-metadata'] = {}
                processed['x-metadata']['xml'] = value
            elif key == 'externalDocs':
                if 'x-metadata' not in processed:
                    processed['x-metadata'] = {}
                processed['x-metadata']['externalDocs'] = value
            elif key == 'example':
                if 'examples' not in processed:
                    processed['examples'] = []
                processed['examples'].append(value)
            elif key == '$ref':
                # Swagger 2.0 refs are already in #/definitions/ format
                processed['$ref'] = value
            elif key in ('properties', 'additionalProperties', 'patternProperties'):
                if key == 'properties' and isinstance(value, dict):
                    processed[key] = {
                        prop_name: self._process_swagger2_schema(prop_schema)
                        for prop_name, prop_schema in value.items()
                    }
                elif key == 'additionalProperties' and isinstance(value, dict):
                    processed[key] = self._process_swagger2_schema(value)
                elif key == 'patternProperties' and isinstance(value, dict):
                    processed[key] = {
                        pattern: self._process_swagger2_schema(prop_schema)
                        for pattern, prop_schema in value.items()
                    }
                else:
                    processed[key] = value
            elif key == 'items':
                if isinstance(value, dict):
                    processed[key] = self._process_swagger2_schema(value)
                elif isinstance(value, list):
                    processed[key] = [self._process_swagger2_schema(item) for item in value]
                else:
                    processed[key] = value
            elif key == 'allOf':
                if isinstance(value, list):
                    processed[key] = [self._process_swagger2_schema(item) for item in value]
                else:
                    processed[key] = value
            elif key == 'type' and value == 'file':
                # Swagger 2.0 file type - convert to string with format
                processed['type'] = 'string'
                processed['format'] = 'binary'
            else:
                processed[key] = value
        
        # Handle nullable conversion
        if processed.get('nullable') is True:
            if 'type' in processed:
                current_type = processed['type']
                if isinstance(current_type, list):
                    if 'null' not in current_type:
                        processed['type'] = current_type + ['null']
                else:
                    processed['type'] = [current_type, 'null']
            del processed['nullable']
        
        return processed
    
    def _lift_inline_schemas_from_swagger2_paths(self, swagger_doc: dict, json_schema: dict) -> None:
        """
        Lift inline schemas from Swagger 2.0 paths into named definitions.
        
        Args:
            swagger_doc: The Swagger 2.0 document.
            json_schema: The JSON Schema being built (modified in place).
        """
        if 'definitions' not in json_schema:
            json_schema['definitions'] = {}
        
        paths = swagger_doc.get('paths', {})
        
        for path, path_item in paths.items():
            if not isinstance(path_item, dict):
                continue
            
            for method in ['get', 'put', 'post', 'delete', 'options', 'head', 'patch']:
                operation = path_item.get(method)
                if not isinstance(operation, dict):
                    continue
                
                operation_id = operation.get('operationId', f"{method}_{path.replace('/', '_')}")
                
                # Extract parameter schemas (body parameters in Swagger 2.0)
                parameters = operation.get('parameters', [])
                for param in parameters:
                    if isinstance(param, dict) and param.get('in') == 'body':
                        schema = param.get('schema', {})
                        if isinstance(schema, dict) and '$ref' not in schema:
                            def_name = f"{operation_id}_Request"
                            processed = self._process_swagger2_schema(schema)
                            json_schema['definitions'][def_name] = processed
                
                # Extract response schemas
                responses = operation.get('responses', {})
                for status_code, response in responses.items():
                    if isinstance(response, dict) and 'schema' in response:
                        schema = response['schema']
                        if isinstance(schema, dict) and '$ref' not in schema:
                            def_name = f"{operation_id}_{status_code}_Response"
                            processed = self._process_swagger2_schema(schema)
                            json_schema['definitions'][def_name] = processed

    def extract_schemas_from_openapi(self, openapi_doc: dict, spec_type: str = None) -> dict:
        """
        Extract schemas from an OpenAPI 3.x document and convert to JSON Schema format.
        
        This method extracts all schema definitions from `components.schemas` and
        creates a consolidated JSON Schema document that can be processed by the
        existing JSON Schema to JSON Structure converter.
        
        Args:
            openapi_doc: The OpenAPI document as a dictionary.
            spec_type: The spec type (OPENAPI_3_0 or OPENAPI_3_1).
            
        Returns:
            A JSON Schema document with definitions from the OpenAPI document.
        """
        # Use appropriate JSON Schema version based on OpenAPI version
        # OpenAPI 3.1 uses JSON Schema draft 2020-12
        if spec_type == self.OPENAPI_3_1:
            schema_version = "https://json-schema.org/draft/2020-12/schema"
        else:
            schema_version = "http://json-schema.org/draft-07/schema#"
        
        json_schema: Dict[str, Any] = {
            "$schema": schema_version
        }
        
        # Copy OpenAPI info to JSON Schema metadata
        if 'info' in openapi_doc:
            info = openapi_doc['info']
            if 'title' in info:
                json_schema['title'] = info['title']
            if 'description' in info:
                json_schema['description'] = info['description']
            if 'version' in info:
                title_slug = info.get('title', 'openapi').lower().replace(' ', '-')
                json_schema['$id'] = f"https://{self.root_namespace}/schemas/{title_slug}/{info['version']}"
        
        # Extract components.schemas
        components = openapi_doc.get('components', {})
        schemas = components.get('schemas', {})
        
        if schemas:
            # Convert OpenAPI schemas to JSON Schema definitions
            json_schema['definitions'] = {}
            for schema_name, schema_def in schemas.items():
                # Process the schema to handle OpenAPI-specific keywords
                processed_schema = self._process_openapi_schema(schema_def)
                json_schema['definitions'][schema_name] = processed_schema
        
        # Optionally lift inline schemas from paths
        if self.lift_inline_schemas:
            self._lift_inline_schemas_from_paths(openapi_doc, json_schema)
        
        return json_schema
    
    def _process_openapi_schema(self, schema: Union[dict, Any]) -> Union[dict, Any]:
        """
        Process an OpenAPI schema to handle OpenAPI-specific keywords.
        
        This method converts OpenAPI-specific keywords to JSON Structure metadata
        annotations or handles them appropriately.
        
        Args:
            schema: The OpenAPI schema object.
            
        Returns:
            The processed schema with OpenAPI keywords handled.
        """
        if not isinstance(schema, dict):
            return schema
        
        processed: Dict[str, Any] = {}
        
        for key, value in schema.items():
            if key == 'nullable':
                # OpenAPI nullable is handled by making type a union with null
                # This is handled during conversion, store it for later processing
                processed['nullable'] = value
            elif key == 'readOnly':
                # Map to JSON Structure metadata annotation
                if 'x-metadata' not in processed:
                    processed['x-metadata'] = {}
                processed['x-metadata']['readOnly'] = value
            elif key == 'writeOnly':
                # Map to JSON Structure metadata annotation
                if 'x-metadata' not in processed:
                    processed['x-metadata'] = {}
                processed['x-metadata']['writeOnly'] = value
            elif key == 'deprecated':
                # Map to JSON Structure metadata annotation
                if 'x-metadata' not in processed:
                    processed['x-metadata'] = {}
                processed['x-metadata']['deprecated'] = value
            elif key == 'discriminator':
                # OpenAPI discriminator - process and convert mapping references
                if isinstance(value, dict):
                    processed_discriminator = dict(value)
                    if 'mapping' in processed_discriminator:
                        # Convert mapping references from OpenAPI to JSON Schema format
                        converted_mapping = {}
                        for key_name, ref_value in processed_discriminator['mapping'].items():
                            if isinstance(ref_value, str) and ref_value.startswith('#/components/schemas/'):
                                converted_mapping[key_name] = ref_value.replace('#/components/schemas/', '#/definitions/')
                            else:
                                converted_mapping[key_name] = ref_value
                        processed_discriminator['mapping'] = converted_mapping
                    processed['discriminator'] = processed_discriminator
                else:
                    processed['discriminator'] = value
            elif key == 'xml':
                # OpenAPI XML object - map to metadata
                if 'x-metadata' not in processed:
                    processed['x-metadata'] = {}
                processed['x-metadata']['xml'] = value
            elif key == 'externalDocs':
                # Map to JSON Structure metadata
                if 'x-metadata' not in processed:
                    processed['x-metadata'] = {}
                processed['x-metadata']['externalDocs'] = value
            elif key == 'example':
                # Map example to JSON Schema examples array
                if 'examples' not in processed:
                    processed['examples'] = []
                processed['examples'].append(value)
            elif key == 'examples':
                # OpenAPI examples object
                processed['examples'] = list(value.values()) if isinstance(value, dict) else value
            elif key == '$ref':
                # Handle OpenAPI references - convert to JSON Schema format
                ref = value
                if ref.startswith('#/components/schemas/'):
                    # Convert OpenAPI ref to JSON Schema ref
                    ref = ref.replace('#/components/schemas/', '#/definitions/')
                processed['$ref'] = ref
            elif key in ('properties', 'additionalProperties', 'patternProperties'):
                # Recursively process nested schemas
                if key == 'properties' and isinstance(value, dict):
                    processed[key] = {
                        prop_name: self._process_openapi_schema(prop_schema)
                        for prop_name, prop_schema in value.items()
                    }
                elif key == 'additionalProperties' and isinstance(value, dict):
                    processed[key] = self._process_openapi_schema(value)
                elif key == 'patternProperties' and isinstance(value, dict):
                    processed[key] = {
                        pattern: self._process_openapi_schema(prop_schema)
                        for pattern, prop_schema in value.items()
                    }
                else:
                    processed[key] = value
            elif key == 'items':
                # Recursively process array items
                if isinstance(value, dict):
                    processed[key] = self._process_openapi_schema(value)
                elif isinstance(value, list):
                    processed[key] = [self._process_openapi_schema(item) for item in value]
                else:
                    processed[key] = value
            elif key in ('allOf', 'anyOf', 'oneOf'):
                # Recursively process composition schemas
                if isinstance(value, list):
                    processed[key] = [self._process_openapi_schema(item) for item in value]
                else:
                    processed[key] = value
            elif key == 'not':
                # Recursively process negation schema
                processed[key] = self._process_openapi_schema(value)
            else:
                # Pass through all other keywords
                processed[key] = value
        
        # Handle nullable by converting to union with null
        if processed.get('nullable') is True:
            if 'type' in processed:
                current_type = processed['type']
                if isinstance(current_type, list):
                    if 'null' not in current_type:
                        processed['type'] = current_type + ['null']
                else:
                    processed['type'] = [current_type, 'null']
            # Remove the nullable keyword after processing
            del processed['nullable']
        elif 'nullable' in processed:
            del processed['nullable']
        
        return processed
    
    def _lift_inline_schemas_from_paths(self, openapi_doc: dict, json_schema: dict) -> None:
        """
        Optionally lift inline schemas from paths/operations into named definitions.
        
        This method extracts inline schemas from request bodies, responses, and parameters
        in the OpenAPI paths section and adds them to the definitions.
        
        Args:
            openapi_doc: The OpenAPI document.
            json_schema: The JSON Schema being built (modified in place).
        """
        if 'definitions' not in json_schema:
            json_schema['definitions'] = {}
        
        paths = openapi_doc.get('paths', {})
        
        for path, path_item in paths.items():
            if not isinstance(path_item, dict):
                continue
            
            for method in ['get', 'put', 'post', 'delete', 'options', 'head', 'patch', 'trace']:
                operation = path_item.get(method)
                if not isinstance(operation, dict):
                    continue
                
                operation_id = operation.get('operationId', f"{method}_{path.replace('/', '_')}")
                
                # Extract request body schemas
                request_body = operation.get('requestBody', {})
                if isinstance(request_body, dict):
                    content = request_body.get('content', {})
                    for media_type, media_type_obj in content.items():
                        if isinstance(media_type_obj, dict) and 'schema' in media_type_obj:
                            schema = media_type_obj['schema']
                            if isinstance(schema, dict) and '$ref' not in schema:
                                # Inline schema - lift to definitions
                                def_name = f"{operation_id}_Request"
                                processed = self._process_openapi_schema(schema)
                                json_schema['definitions'][def_name] = processed
                
                # Extract response schemas
                responses = operation.get('responses', {})
                for status_code, response in responses.items():
                    if isinstance(response, dict):
                        content = response.get('content', {})
                        for media_type, media_type_obj in content.items():
                            if isinstance(media_type_obj, dict) and 'schema' in media_type_obj:
                                schema = media_type_obj['schema']
                                if isinstance(schema, dict) and '$ref' not in schema:
                                    # Inline schema - lift to definitions
                                    def_name = f"{operation_id}_{status_code}_Response"
                                    processed = self._process_openapi_schema(schema)
                                    json_schema['definitions'][def_name] = processed
    
    def convert_openapi_to_structure(
        self, 
        openapi_doc: Union[dict, str], 
        base_uri: str = ''
    ) -> dict:
        """
        Convert an OpenAPI/Swagger document to JSON Structure format.
        
        Supports:
        - Swagger 2.0
        - OpenAPI 3.0.x
        - OpenAPI 3.1.x
        
        Args:
            openapi_doc: The API document as a dictionary or JSON string.
            base_uri: The base URI for resolving references.
            
        Returns:
            The JSON Structure document.
            
        Raises:
            ValueError: If the input is not a valid Swagger/OpenAPI document.
            TypeError: If the input type is not supported.
        """
        # Parse JSON string if needed
        if isinstance(openapi_doc, str):
            openapi_doc = json.loads(openapi_doc)
        
        if not isinstance(openapi_doc, dict):
            raise TypeError(f"Expected dict or str, got {type(openapi_doc)}")
        
        # Detect and validate specification version
        spec_type, version = self.detect_spec_version(openapi_doc)
        
        # Extract schemas based on specification type
        if spec_type == self.SWAGGER_2:
            json_schema = self.extract_schemas_from_swagger2(openapi_doc)
        else:
            json_schema = self.extract_schemas_from_openapi(openapi_doc, spec_type)
        
        # Use the JSON Schema to JSON Structure converter
        json_converter = JsonToStructureConverter()
        json_converter.root_namespace = self.root_namespace
        json_converter.root_class_name = self.root_class_name
        json_converter.preserve_composition = self.preserve_composition
        json_converter.detect_inheritance = self.detect_inheritance
        json_converter.detect_discriminators = self.detect_discriminators
        json_converter.convert_empty_objects_to_maps = self.convert_empty_objects_to_maps
        
        # Convert to JSON Structure
        structure_schema = json_converter.convert_json_schema_to_structure(json_schema, base_uri)
        
        return structure_schema
    
    def convert_openapi_file_to_structure(
        self,
        openapi_file_path: str,
        structure_file_path: Optional[str] = None
    ) -> dict:
        """
        Convert an OpenAPI file to JSON Structure format.
        
        Args:
            openapi_file_path: Path to the input OpenAPI file.
            structure_file_path: Optional path for the output JSON Structure file.
            
        Returns:
            The JSON Structure document.
        """
        # Read the OpenAPI file
        with open(openapi_file_path, 'r', encoding='utf-8') as f:
            content = f.read()
        
        # Detect format (JSON or YAML)
        try:
            openapi_doc = json.loads(content)
        except json.JSONDecodeError:
            # Try YAML
            try:
                import yaml
                openapi_doc = yaml.safe_load(content)
            except ImportError:
                raise ValueError("YAML support requires PyYAML. Install with: pip install pyyaml")
            except Exception as e:
                raise ValueError(f"Failed to parse OpenAPI document as JSON or YAML: {e}")
        
        # Convert to JSON Structure
        base_uri = f"file://{os.path.abspath(openapi_file_path)}"
        structure_schema = self.convert_openapi_to_structure(openapi_doc, base_uri)
        
        # Write output if path specified
        if structure_file_path:
            with open(structure_file_path, 'w', encoding='utf-8') as f:
                json.dump(structure_schema, f, indent=2)
        
        return structure_schema


def convert_openapi_to_structure(
    input_data: str,
    root_namespace: str = 'example.com'
) -> str:
    """
    Convert an OpenAPI/Swagger document to JSON Structure format.
    
    Supports Swagger 2.0, OpenAPI 3.0.x, and OpenAPI 3.1.x.
    
    Args:
        input_data: The API document as a JSON string.
        root_namespace: The namespace for the root schema.
        
    Returns:
        The JSON Structure document as a JSON string.
    """
    converter = OpenApiToStructureConverter()
    converter.root_namespace = root_namespace
    
    result = converter.convert_openapi_to_structure(input_data)
    
    return json.dumps(result, indent=2)


def convert_openapi_to_structure_files(
    openapi_file_path: str,
    structure_schema_path: str,
    root_namespace: Optional[str] = None,
    preserve_composition: bool = False,
    detect_discriminators: bool = True,
    lift_inline_schemas: bool = False
) -> None:
    """
    Convert an OpenAPI/Swagger file to JSON Structure format.
    
    Supports Swagger 2.0, OpenAPI 3.0.x, and OpenAPI 3.1.x.
    
    Args:
        openapi_file_path: Path to the input API file (JSON or YAML).
        structure_schema_path: Path to the output JSON Structure file.
        root_namespace: The namespace for the root schema.
        preserve_composition: Flag to preserve composition keywords.
        detect_discriminators: Flag to detect discriminator patterns.
        lift_inline_schemas: Flag to lift inline schemas from paths to definitions.
    """
    if root_namespace is None:
        root_namespace = 'example.com'
    
    converter = OpenApiToStructureConverter()
    converter.root_namespace = root_namespace
    converter.preserve_composition = preserve_composition
    converter.detect_discriminators = detect_discriminators
    converter.lift_inline_schemas = lift_inline_schemas
    
    converter.convert_openapi_file_to_structure(openapi_file_path, structure_schema_path)



"""
Module to convert Protobuf .proto files to Avro schema.
"""

import json
import os
import re
from typing import Dict, List, Tuple
from avrotize.common import pascal
from avrotize.dependency_resolver import sort_messages_by_dependencies, inline_dependencies_of
from . import proto2parser
from . import proto3parser

AvroSchema = Dict[str, 'AvroSchema'] | List['AvroSchema'] | str | None

class ProtoToAvroConverter:
    """Class to convert Protobuf .proto files to Avro schema."""

    isomorphic_types = ['float', 'double', 'bytes', 'string']

    def __init__(self, proto_root: str = None):
        """Initialize ProtoToAvroConverter.
        
        Args:
            proto_root (str): Optional root directory for resolving proto imports. 
                             When provided, imports are resolved relative to this directory.
        """
        self.imported_types: Dict[str, str] = {}
        self.generated_types: Dict[str, str] = {}
        self.forward_references: Dict[str, str] = {} # table for resolvbing forward references
        self.proto_root: str = proto_root

    def proto_type_to_avro_primitive(self, proto_type: str)-> Tuple[bool, str]:
        """
        Map Protobuf types to Avro primitive types.

        Args:
            proto_type (str): Protobuf type to convert.

        Returns:
            str or dict: Corresponding Avro type.
        """
        mapping = {
            'google.protobuf.Empty': 'null',  # Special handling may be required
            'bool': 'boolean',
            'int32': 'int',
            'uint32': 'int',
            'sint32': 'int',
            'int64': 'long',
            'uint64': 'long',
            'sint64': 'long',
            'fixed32': 'int',
            'fixed64': 'long',
            'sfixed32': 'int',
            'sfixed64': 'long',
            'google.protobuf.Timestamp': {
                "type": "long",
                "logicalType": "timestamp-micros"
            }
        }
        if proto_type in self.isomorphic_types:
            return True, proto_type
        mapped = mapping.get(proto_type, None)
        if mapped:
            return True, mapped
        return False, proto_type

    def build_forward_references_from_message(self, proto_message_type: proto2parser.Message | proto3parser.Message, avro_namespace: str):
        """
        Build forward references from a Protobuf message.

        Args:
            proto_message_type: The message type from the parsed proto file.
            avro_namespace (str): The namespace for the message.
        """
        for _, nested_message in proto_message_type.messages.items():
            nested_namespace = avro_namespace + '.' + proto_message_type.name + '_types'
            self.build_forward_references_from_message(nested_message, nested_namespace)
        for _, enum_type in proto_message_type.enums.items():
            nested_namespace = avro_namespace + '.' + proto_message_type.name + '_types'
            self.forward_references[nested_namespace+'.'+enum_type.name] = "enum"
        self.forward_references[avro_namespace+'.'+proto_message_type.name] = "record"

    def build_forward_references_from_file(self, proto_file:  proto3parser.ProtoFile| proto2parser.ProtoFile, avro_namespace: str):
        """
        Build forward references from a Protobuf file.

        Args:
            proto_file: The parsed proto file.
            avro_namespace (str): The namespace for the message.
        """
        for _, enum_type in proto_file.enums.items():
             self.forward_references[avro_namespace+'.'+enum_type.name] = "enum"
        for _, message in proto_file.messages.items():
            self.build_forward_references_from_message(message, avro_namespace)

    def convert_proto_to_avro_schema(self, proto_file_path: str, avro_namespace: str, message_type: str) -> list:
        """
        Convert .proto file to Avro schema.

        Args:
            proto_file_path (str): Path to the Protobuf .proto file.

        Returns:
            list: Avro schema as a list of dictionaries.
        """
        with open(proto_file_path, 'r', encoding='utf-8') as proto_file:
            proto_schema = proto_file.read()

        # Determine whether we have proto3 or proto2 and parse the data
        if re.search(r'syntax\s*=\s*"proto3"', proto_schema):
            data: proto3parser.ProtoFile = proto3parser.parse(proto_schema)
        else:
            data: proto2parser.ProtoFile = proto2parser.parse(proto_schema)

        # Build forward references
        self.build_forward_references_from_file(data, avro_namespace)
        # Avro schema header
        avro_schema = []
        for import_ in data.imports:
            # Handle protobuf imports
            if import_.startswith('google/protobuf/'):
                script_path = os.path.dirname(os.path.abspath(__file__))
                avsc_dir = os.path.join(script_path, 'prototypes')
                # Load the corresponding avsc file from ./prototypes at this script's path into avro_schema
                avsc = f'{avsc_dir}/{import_.replace("google/protobuf/", "").replace(".proto", ".avsc")}'
                with open(avsc, 'r', encoding='utf-8') as avsc_file:
                    types = json.load(avsc_file)
                    for t in types:
                        qualified_name = t["namespace"] + "." + t["name"]
                        self.imported_types[qualified_name] = t
            else:
                # Resolve import path: try proto_root first, then fall back to file-relative path
                import_path = None
                
                if self.proto_root:
                    # Try resolving relative to proto_root
                    candidate_path = os.path.join(self.proto_root, import_)
                    if os.path.exists(candidate_path):
                        import_path = candidate_path
                
                if not import_path:
                    # Fall back to resolving relative to the directory of the current proto file
                    cwd = os.path.join(os.getcwd(), os.path.dirname(proto_file_path))
                    candidate_path = os.path.join(cwd, import_)
                    if os.path.exists(candidate_path):
                        import_path = candidate_path
                
                # Raise an exception if the imported file does not exist
                if not import_path:
                    raise FileNotFoundError(f'Import file {import_} does not exist. Searched in proto_root: {self.proto_root}, and relative to: {os.path.dirname(proto_file_path)}')
                
                package_name = pascal(import_.replace('.proto', ''))
                import_namespace = (avro_namespace + '.' + package_name) if avro_namespace else package_name
                avro_schema.extend(self.convert_proto_to_avro_schema(import_path, import_namespace, message_type))


        # Convert enum fields
        for _, enum_type in data.enums.items():
            self.handle_enum(enum_type, avro_schema, avro_namespace)

        # Convert message fields
        for _, m in data.messages.items():
            self.handle_message(m, avro_schema, avro_namespace)


        # Sort the messages in avro_schema by dependencies
        if message_type:
            message_schema = next(
                (message for message in avro_schema if message['type'] == "record" and message['name'] == message_type), None)
            if not message_schema:
                raise ValueError(f'Message type {message_type} not found in the Avro schema.')
            else:
                inline_dependencies_of(avro_schema, message_schema)
                return message_schema
        else:
            avro_schema = sort_messages_by_dependencies(avro_schema)
        return avro_schema

    @staticmethod
    def clean_comment(comment: str):
        """
        Clean comments by stripping slashes, newlines, linefeeds, and extra whitespace.

        Args:
            comment (str): The comment to clean.

        Returns:
            str: Cleaned comment.
        """
        if comment:
            return comment.replace('//', '').replace('\n', '').lstrip().rstrip()
        return None

    def handle_enum(self, enum_type: proto2parser.Enum | proto3parser.Enum, avro_schema: AvroSchema, avro_namespace: str) -> AvroSchema:
        """
        Convert enum fields to avro schema.

        Args:
            enum_type: The enum type from the parsed proto file.
            avro_schema (list): The list to append the converted enum schema.
            namespace (str): The namespace for the enum.
        """
        comment = self.clean_comment(
            enum_type.comment.content if enum_type.comment and enum_type.comment.content else None)

        # Create avro schema
        avro_enum: AvroSchema = {
            'name': enum_type.name,
            'type': 'enum',
            'namespace': avro_namespace,
            'symbols': [],
            'ordinals': {}
        }

        if comment:
            avro_enum['doc'] = comment
        for value in enum_type.fields:
            avro_enum['symbols'].append(value.name)
            avro_enum['ordinals'][value.name] = int(value.number)
        avro_schema.append(avro_enum)
        self.generated_types[avro_enum['namespace']+'.'+avro_enum['name']] = "enum"
        return avro_enum

    def handle_message(self, proto_message_type: proto2parser.Message | proto3parser.Message, avro_schema: AvroSchema, avro_namespace: str)-> AvroSchema:
        """
        Convert protobuf messages to avro records.

        Args:
            m: The message type from the parsed proto file.
            avro_schema (list): The list to append the converted message schema.
            namespace (str): The namespace for the message.
        """
        dependencies = []

        comment = self.clean_comment(proto_message_type.comment.content if proto_message_type.comment and proto_message_type.comment.content else None)
        avro_record: AvroSchema = {
            'type': 'record',
            'name': proto_message_type.name,
            'namespace': avro_namespace,
            'fields': []
        }
        if comment:
            avro_record['doc'] = comment
        for proto_field in proto_message_type.fields:
            avro_type = self.get_avro_type_for_field(proto_message_type, avro_namespace, avro_schema, dependencies, proto_field)
            comment = self.clean_comment(proto_field.comment.content if proto_field.comment and proto_field.comment.content else None)

            avro_field = {
                'name': proto_field.name,
                'type': avro_type,
            }

            if comment:
                avro_field['doc'] = comment

            avro_record['fields'].append(avro_field)

        for proto_field in proto_message_type.oneofs:
            avro_oneof: AvroSchema = {
                'name': proto_field.name,
                'type': []
            }
            comment = self.clean_comment(proto_field.comment.content if proto_field.comment and proto_field.comment.content else None)
            if comment:
                avro_oneof['doc'] = comment
            for oneof_field in proto_field.fields:
                avro_type = self.get_avro_type_for_field(proto_message_type, avro_namespace, avro_schema, dependencies, oneof_field)
                comment = self.clean_comment(oneof_field.comment.content if oneof_field.comment and oneof_field.comment.content else None)
                if comment:
                    oneof_field['doc'] = comment
                avro_oneof['type'].append(avro_type)
            avro_record['fields'].append(avro_oneof)

        if dependencies:
            avro_record['dependencies'] = dependencies
        avro_schema.append(avro_record)
        for _, nested_message in proto_message_type.messages.items():
            nested_namespace = avro_namespace + '.' + proto_message_type.name + '_types'
            self.handle_message(nested_message, avro_schema, nested_namespace)
        # Convert enum fields
        for _, enum_type in proto_message_type.enums.items():
            nested_namespace = avro_namespace + '.' + proto_message_type.name + '_types'
            self.handle_enum(enum_type, avro_schema, nested_namespace)
        self.generated_types[avro_record['namespace']+'.'+avro_record['name']] = "record"
        return avro_record

    def get_avro_type_for_field(self, proto_message_type: proto2parser.Message | proto3parser.Message, avro_namespace: str, avro_schema: AvroSchema, dependencies: List[str], proto_field: proto2parser.Field | proto3parser.Field):
        """
        Get Avro type for a Protobuf field.

        Args:
            m: The message type from the parsed proto file.
            namespace (str): The namespace for the message.
            dependencies (list): The list to append the dependencies.
            f: The field from the parsed proto file.

        Returns:
            str or dict: Corresponding Avro type.
        """
        avro_field_type: AvroSchema = None
        proto_field_type = proto_field.val_type if proto_field.label == 'repeated' or proto_field.type == 'map' else proto_field.type
        is_primitive, avro_field_type = self.proto_type_to_avro_primitive(proto_field_type)

        if not is_primitive:
            if proto_field.type in self.imported_types:
                avro_field_type = self.imported_types[proto_field.type]
            else:
                avro_field_type = avro_namespace + '.' + avro_field_type
                found_in_nested_definitions = False
                for k, nested_proto_message_type in proto_message_type.messages.items():
                    nested_namespace = avro_namespace + '.' + proto_message_type.name + '_types'
                    if nested_proto_message_type.name == proto_field_type:
                        avro_field_type = self.handle_message(nested_proto_message_type, avro_schema, nested_namespace)
                        del proto_message_type.messages[k]
                        if 'dependencies' in avro_field_type:
                            dependencies.extend(avro_field_type['dependencies'])
                            del avro_field_type['dependencies']
                        found_in_nested_definitions = True
                        break
                if not found_in_nested_definitions:
                    for k, nested_proto_enum_type in proto_message_type.enums.items():
                        nested_namespace = avro_namespace + '.' + proto_message_type.name + '_types'
                        if nested_proto_enum_type.name == proto_field_type:
                            avro_field_type = self.handle_enum(nested_proto_enum_type, avro_schema, nested_namespace)
                            del proto_message_type.enums[k]
                            found_in_nested_definitions = True
                            break
                if not found_in_nested_definitions:
                    dependency_avro_field_type = avro_field_type
                    while '.' in dependency_avro_field_type:
                        if dependency_avro_field_type in self.forward_references:
                            dependencies.append(dependency_avro_field_type)
                            break
                        n = dependency_avro_field_type.split('.')
                        dependency_avro_field_type = '.'.join(n[:-2]+[n[-1]])

        if proto_field.label == 'optional':
            avro_field_type = ["null", avro_field_type]
        if proto_field.label == 'repeated':
            avro_type: AvroSchema = {
                "type": "array",
                "items": avro_field_type
            }
        elif proto_field.type == 'map':
            avro_type: AvroSchema = {
                "type": "map",
                "values": avro_field_type,
            }
        else:
            avro_type = avro_field_type
        return avro_type


def convert_proto_to_avro(proto_file_path: str, avro_schema_path: str, namespace: str = None, message_type: str = None, proto_root: str = None):
    """
    Convert Protobuf .proto file to Avro schema.

    Args:
        proto_file_path (str): Path to the Protobuf .proto file.
        avro_schema_path (str): Path to save the Avro schema .avsc file.
        namespace (str): Optional namespace for the Avro schema.
        message_type (str): Optional specific message type to extract.
        proto_root (str): Optional root directory for resolving proto imports.
                         When provided, imports are resolved relative to this directory.

    Raises:
        FileNotFoundError: If the proto file does not exist.
        ValueError: If the file extensions are incorrect.
    """
    if not os.path.exists(proto_file_path):
        raise FileNotFoundError(f'Proto file {proto_file_path} does not exist.')

    converter = ProtoToAvroConverter(proto_root=proto_root)
    if not namespace:
        namespace = pascal(os.path.basename(proto_file_path).replace('.proto', ''))
    avro_schema = converter.convert_proto_to_avro_schema(proto_file_path, namespace, message_type)

    # Convert the Avro schema to JSON and write it to the file
    with open(avro_schema_path, 'w', encoding='utf-8') as avro_file:
        avro_file.write(json.dumps(avro_schema, indent=2))

